/*
 *
 * This program is written by
 * Vincent Guerin
 * and Ludovic Rousseau <ludovic.rousseau@gemplus.com>
 *
 * Code changed to use PCSC instead of TLP calls
 * November 2001 - Domenico Taddei <domenico.taddei@proteam-group.it>
 *
 * $Id: smartcard_functions.c,v 1.16 2002/08/19 08:47:00 rousseau Exp $
 *
 */

/*
 * Copyright  2000 Gemplus
 * Av. du Pic de Bertagne - Parc d'Activits de Gmenos
 * BP 100 - 13881 Gmenos CEDEX
 * 
 * Everyone is allowed to redistribute and use this source  (source
 * code)  and binary (object code),  with or  without modification,
 * under some conditions:
 * 
 *  - Everyone  must  retain  and/or  reproduce the above copyright
 *    notice,  and the below  disclaimer of warranty and limitation
 *    of liability  for redistribution and use of these source code
 *    and object code.
 * 
 *  - Everyone  must  ask a  specific prior written permission from
 *    Gemplus to use the name of Gemplus.
 * 
 * DISCLAIMER OF WARRANTY
 * 
 * THIS CODE IS PROVIDED "AS IS",  WITHOUT ANY WARRANTY OF ANY KIND
 * (INCLUDING,  BUT  NOT  LIMITED  TO,  THE IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY  AND FITNESS FOR  A  PARTICULAR PURPOSE)  EITHER
 * EXPRESS OR IMPLIED.  GEMPLUS DOES NOT WARRANT THAT THE FUNCTIONS
 * CONTAINED  IN THIS SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR
 * THAT THE OPERATION OF IT WILL BE UNINTERRUPTED OR ERROR-FREE. NO
 * USE  OF  ANY  CODE  IS  AUTHORIZED  HEREUNDER EXCEPT UNDER  THIS
 * DISCLAIMER.
 * 
 * LIMITATION OF LIABILITY
 * 
 * GEMPLUS SHALL NOT BE LIABLE FOR INFRINGEMENTS OF  THIRD  PARTIES
 * RIGHTS. IN NO EVENTS, UNLESS REQUIRED BY APPLICABLE  LAW,  SHALL
 * GEMPLUS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES OF ANY CHARACTER  INCLUDING,
 * WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK STOPPAGE,
 * COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER DAMAGES OR
 * LOSSES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. ALSO,
 * GEMPLUS IS  UNDER NO  OBLIGATION TO MAINTAIN,  CORRECT,  UPDATE,
 * CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS SOFTWARE.
 */
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>
#include <winscard.h>

#include "config.h"
#include "debug_config.h"
#include "debug.h"
#include "global.h"
#include "apdu.h"
#include "pam_smartcard_utils.h"
#include "smartcard_functions.h"
#include "publickey_db_functions.h"
#include "crypto_functions.h"

// These functions works on the smartcard.
//////////////////////////////////////////

// define BUGGY_GPK_KEY_GENERATION if you have a card older than GPK8000 v3
#define BUGGY_GPK_KEY_GENERATION
#define BUGGY_GPK_KEY_GENERATION_TIME 20

// These constants are used in the admninistrator PIN code
// generation in order to diversify to generated codes.
#define ADM_PIN_NR 1
#define ADM_PUC_NR 2
#define CERT_ADM_PIN_NR 3
#define CERT_ADM_PUC_NR 4


// Global

// Internal functions
/////////////////////

// Generates administor PIN and PUC code using the passphrase and
// the card serial number.
void generate_code(char *code, char *passphrase, int passphraseLen,
	char *csn, unsigned char nrCode, int codeLen);

// Creates a Dedicated file on the card using the secure messaging
// mechanism.
int create_df_with_secure_messaging(char *DF_name, int DF_nameLen,
	unsigned char DF_sfi);

// This function creates and initializes a RSA key file on the card.
// It also manages the on-board generation of the pair of keys.
int initialize_and_generate_RSA_key(int rsaKeySize,
	unsigned char protectionMode, unsigned char codeNumber,
	unsigned char rsaKeySfi);

// This function returns the card serial number
int get_card_serial_number(char *csn);

// This function selects the specified Dedicated File
int select_DF(unsigned char DF_SFI, int quiet);

#ifdef DEBUG_LEVEL_INFO

static struct timeval start_tv;

#define DEBUG_MSG(a) \
{ \
	struct timeval tv; \
	\
	gettimeofday(&tv, NULL); \
	tv.tv_sec -= start_tv.tv_sec; \
	tv.tv_usec -= start_tv.tv_usec; \
	if (tv.tv_usec<0) \
	{ \
		tv.tv_usec += 1000000; \
		tv.tv_sec -= 1; \
	} \
	fprintf(stderr, "%ld %06ld: %s\n", tv.tv_sec, tv.tv_usec, a); \
}

#define DEBUG_INIT \
{ \
	gettimeofday(&start_tv, NULL); \
}

#else
#define DEBUG_INIT
#define DEBUG_MSG(a)
#endif


int change_pin(char *currentPin, char *newPin)
{
	int rCode, i;

	BYTE recv[BUF_SIZE];
	BYTE apdu[BUF_SIZE];
	DWORD recvlen;
	SCARD_IO_REQUEST pio_recv_pci;

	// Card connection
	rCode = pcsc_init(0);

	if (rCode != SCARD_S_SUCCESS)
	{
		pcsc_release();
		return (rCode);
	}

	// Select the PAM Dedicated File
	rCode = select_DF(PAM_DF_SFI, 0);

	if (rCode == EXIT_FAILURE)
		return (ERROR_CARD_NOT_FORMATED);

	// Send the change PIN instruction
	//////////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = SET_SECRET_CODE;
	apdu[2] = P1_CHANGE_SECRET_CODE;
	apdu[3] = USR_PIN_CODE_NUMBER;
	apdu[4] = SECRET_CODE_SIZE;

	// Old PIN Value
	for (i = 0; i < USR_PIN_SIZE / 2; i++)
	{
		int tmp;
		sscanf(currentPin + 2 * i, "%02x", &tmp);
		apdu[5 + i] = (unsigned char) tmp;
	}

	// Padd the end of the PIN with 0
	for (i = USR_PIN_SIZE / 2; i < SECRET_CODE_SIZE / 2; i++)
		apdu[5 + i] = 0x00;

	// New PIN Value
	for (i = 0; i < USR_PIN_SIZE / 2; i++)
	{
		int tmp;
		sscanf(newPin + 2 * i, "%02x", &tmp);
		apdu[5 + i + SECRET_CODE_SIZE / 2] = (unsigned char) tmp;
	}

	// Padd the end of the PIN with 0
	for (i = USR_PIN_SIZE / 2; i < SECRET_CODE_SIZE / 2; i++)
		apdu[5 + i + SECRET_CODE_SIZE / 2] = 0x00;
	recvlen = 2;
	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5 + SECRET_CODE_SIZE,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Change User PIN Code: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Changing User PIN Code: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		if (recv[recvlen - 2] == SW1_PIN_ERROR)
			return (ERROR_VERIFY_PIN);
		// We distinguish the specific wrong PIN error 
		else
			return (EXIT_FAILURE);
	}

	// Disconnection
	pcsc_release();

	return (EXIT_SUCCESS);
}


int get_cfs_secret(char *userPin, unsigned char *cfsSecret, int quiet)
{
	// This function retrieves the CFS secret store on the card
	// and stores it in cfsSecret. If quiet flag is set, be silent.
	///////////////////////////////////////////////////////////////

	int rCode, i;

	BYTE recv[BUF_SIZE];
	BYTE apdu[BUF_SIZE];
	DWORD recvlen;
	SCARD_IO_REQUEST pio_recv_pci;

	// Card connection
	DEBUG_INIT;
	DEBUG_MSG("Card connection");
	rCode = pcsc_init(0);

	if (rCode != SCARD_S_SUCCESS)
	{
		pcsc_release();
		return (rCode);
	}

	// Select the PAM Dedicated File
	DEBUG_MSG("Select the PAM Dedicated File");
	rCode = select_DF(PAM_DF_SFI, quiet);

	if (rCode == EXIT_FAILURE)
		return (ERROR_CARD_NOT_FORMATED);

	// PIN checking in order to allow the reading of the CFS secret file
	DEBUG_MSG("PIN checking");
	////////////////////////////////////////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = VERIFY;
	apdu[2] = P1_VERIFY;
	apdu[3] = USR_PIN_CODE_NUMBER;
	apdu[4] = SECRET_CODE_SIZE;

	// Copy the PIN in the data buffer
	for (i = 0; i < USR_PIN_SIZE; i++)
		apdu[5 + i] = userPin[i] - '0';

	// Padd the end of the pin with 0
	for (i = USR_PIN_SIZE; i < SECRET_CODE_SIZE; i++)
		apdu[5 + i] = '0';
	recvlen = 2;
	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5 + SECRET_CODE_SIZE,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		if (!quiet)
		{
			printf("PC/SC Error Sending APDU Verify User PIN Code: %s\n",
				pcsc_stringify_error(rCode));
		};
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		if (!quiet)
		{
			printf("Error Verifying User PIN Code: %s\n",
				ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		};
		pcsc_release();
		if (recv[recvlen - 2] == SW1_PIN_ERROR)
			return (ERROR_VERIFY_PIN);
		// We distinguish the specific wrong PIN error 
		else
			return (EXIT_FAILURE);
	}

	// Get the public CFS secret by reading the corresponding file
	DEBUG_MSG("Get the public CFS secret");
	//////////////////////////////////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = READ_BINARY;
	apdu[2] = P1_SELECT_READ_FROM_EF | CFS_SECRET_SFI;
	apdu[3] = P2_OFFSET_NULL;
	apdu[4] = CFS_SECRET_SIZE;

	recvlen = CFS_SECRET_SIZE + 2;
	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		if (!quiet)
		{
			printf("PC/SC Error Sending APDU Read CFS Secret File: %s\n",
				pcsc_stringify_error(rCode));
		};
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		if (!quiet)
		{
			printf("Error Reading CFS Secret File: %s\n",
				ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		};
		pcsc_release();
		return (EXIT_FAILURE);
	}

	memcpy((unsigned char *) cfsSecret, recv, CFS_SECRET_SIZE);

	// Card disconnection
	DEBUG_MSG("Card disconnection");
	pcsc_release();

	return (EXIT_SUCCESS);
}


int restore_cfs_secret(char *passphrase, int passphraseLen,
	unsigned char *cipheredCfsSecret, unsigned char *cfsSecret)
{
	// Uncipher the ciphered CFS secret using the administrator
	// certification smartcard
	///////////////////////////////////////////////////////////

	int rCode, i;

	BYTE recv[BUF_SIZE];
	BYTE apdu[BUF_SIZE];
	DWORD recvlen;
	SCARD_IO_REQUEST pio_recv_pci;

	char csn[CSN_SIZE];
	char admPin[SECRET_CODE_SIZE];
	unsigned char invertCfsSecret[CFS_SECRET_SIZE];
	unsigned char invertCipheredCfsSecret[CFS_SECRET_SIZE];

	// Card connection
	rCode = pcsc_init(0);

	if (rCode != SCARD_S_SUCCESS)
	{
		pcsc_release();
		return (rCode);
	}

	// Get the card serial number
	rCode = get_card_serial_number(csn);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Generate administrator PIN code
	generate_code(admPin, passphrase, passphraseLen, csn, CERT_ADM_PIN_NR,
		ADM_PIN_SIZE);

	// Select CERT Dedicated File
	rCode = select_DF(CERT_DF_SFI, 0);

	if (rCode == EXIT_FAILURE)
		return (ERROR_CARD_NOT_FORMATED);

	// Verify the administrator PIN code to gain
	// usage of the private key
	////////////////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = VERIFY;
	apdu[2] = P1_VERIFY;
	apdu[3] = CERT_ADM_PIN_CODE_NUMBER;
	apdu[4] = SECRET_CODE_SIZE;

	// Copy the PIN in the data buffer
	for (i = 0; i < ADM_PIN_SIZE; i++)
		apdu[5 + i] = admPin[i] - '0';

	// Padd the end of the PIN with 0
	for (i = ADM_PIN_SIZE; i < SECRET_CODE_SIZE; i++)
		apdu[5 + i] = '0';
	recvlen = 2;
	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5 + SECRET_CODE_SIZE,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU verify administrator PIN: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Verifying administrator PIN: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Select the RSA private key file 
	//////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = SELECT_FILE;
	apdu[2] = P1_SELECT_EF_UNDER_CURRENT_DF;
	apdu[3] = P2_NO_REPONSE_REQUIRED;
	apdu[4] = SFI_SIZE;

	// Fill up the data buffer
	apdu[5] = CERT_DF_SFI;
	apdu[6] = REST_RSA_KEY_SFI;
	recvlen = 2;
	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 7, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Select RSA Private Key File: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Selecting RSA Private Key File: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Select Crypto Context : RSA Unwrap 
	/////////////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = SELECT_CRYPTO_CONTEXT;
	apdu[2] = REST_RSA_KEY_SFI;
	apdu[3] = P2_MODE_RSA_UNWRAP;
	apdu[4] = NULL_BYTE;

	// No data
	recvlen = 2;
	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Select Crypto Context: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Selecting Crypto Context: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Send the ciphered CFS secret to the card for unwrapping.
	////////////////////////////////////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = PK_OPEN_ENVELOPE;
	apdu[2] = NULL_BYTE;
	apdu[3] = P2_RAW_ENVELOPE;
	apdu[4] = KEY_SIZE_1024;

	// Invert the secret
	for (i = 0; i < CFS_SECRET_SIZE; i++)
		invertCipheredCfsSecret[CFS_SECRET_SIZE - i - 1] =
			cipheredCfsSecret[i];

	memcpy(&apdu[5], invertCipheredCfsSecret, CFS_SECRET_SIZE);
	recvlen = 2;
	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5 + CFS_SECRET_SIZE,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Unwrap the CFS secret: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Unwraping the CFS secret: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Get The unwrapped CFS secret
	////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = GET_RESPONSE;
	apdu[2] = NULL_BYTE;
	apdu[3] = NULL_BYTE;
	apdu[4] = KEY_SIZE_1024;
	recvlen = KEY_SIZE_1024 + 2;
	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Get the Unwrapped CFS Secret: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error  Getting the Unwrapped CFS Secret: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	memcpy(invertCfsSecret, recv, KEY_SIZE_1024);

	pcsc_release();

	// Invert the secret  
	for (i = 0; i < CFS_SECRET_SIZE; i++)
		cfsSecret[i] = invertCfsSecret[CFS_SECRET_SIZE - i - 1];

	return (EXIT_SUCCESS);
}

//--------------------------->>>>UNCHANGED<<<--------------------------

void generate_code(char *code, char *passphrase, int passphraseLen,
	char *csn, unsigned char nrCode, int codeLen)
{
	int i;
	char *buffer, hashed[16];
	int len;
	struct MD5Context md5context;

	len = passphraseLen + CSN_SIZE + sizeof(nrCode) + 1;
	buffer = calloc(len, sizeof(char));

	strncpy(buffer, passphrase, passphraseLen);
	strncpy(buffer + passphraseLen, csn, CSN_SIZE);
	strncpy(buffer + passphraseLen + CSN_SIZE, (char *) &nrCode,
		sizeof(nrCode));

	// Hash the concatenation of the card serial number the passprase 
	// and the number of the code in order to generate a unique code

	MD5Init(&md5context);
	MD5Update(&md5context, buffer, len - 1);
	MD5Final(hashed, &md5context);

	// Copy the required number of bytes in the code

	for (i = 0; i < codeLen; i++)
		code[i] = abs(hashed[i] % 10) + '0';

	// Padd with 0 

	for (i = codeLen; i < SECRET_CODE_SIZE; i++)
		code[i] = '0';

	free(buffer);
}

//------------------------------>>>>CHANGED<<<<---------------------------------

int formatcard(char *passphrase, int passphraseLen)
{
	int rCode, i, fd;

	BYTE recv[BUF_SIZE];
	BYTE apdu[BUF_SIZE];
	DWORD recvlen;
	SCARD_IO_REQUEST pio_recv_pci;

	char csn[CSN_SIZE];
	char admPin[SECRET_CODE_SIZE];
	char admPuc[SECRET_CODE_SIZE];
	//unsigned char cfsSecret[CFS_SECRET_SIZE + 1];

	// Card connection
	DEBUG_INIT;
	DEBUG_MSG("Card connection");
	rCode = pcsc_init(0);

	if (rCode != SCARD_S_SUCCESS)
	{
		pcsc_release();
		return (rCode);
	}

	// Get the card serial number
	DEBUG_MSG("Get the card serial number");


	rCode = get_card_serial_number(csn);

	if (rCode != SCARD_S_SUCCESS)
	{
		pcsc_release();
		return (rCode);
	}

	// Generate administrator PIN and PUC codes
	DEBUG_MSG("Generate administrator PIN and PUC codes");
	generate_code(admPin, passphrase, passphraseLen, csn, ADM_PIN_NR,
		ADM_PIN_SIZE);
	generate_code(admPuc, passphrase, passphraseLen, csn, ADM_PUC_NR,
		ADM_PUC_SIZE);

	// Create the PAM Dedicated File
	DEBUG_MSG("Create the PAM Dedicated File");
	rCode =
		create_df_with_secure_messaging(PAM_DF_NAME, PAM_DF_NAME_LEN,
		PAM_DF_SFI);
	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Select the PAM Dedicated File
	DEBUG_MSG("Select the PAM Dedicated File");
	rCode = select_DF(PAM_DF_SFI, 0);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Create the secret codes file
	DEBUG_MSG("Create the secret codes file");
	///////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = CREATE_FILE;
	apdu[2] = EF_TYPE;
	apdu[3] = NULL_BYTE;
	apdu[4] = CREATE_EF_DATA_SIZE;

	// Fill up the data buffer

	// SFI
	apdu[5] = PAM_DF_SFI;
	apdu[6] = SECRET_CODE_SFI;

	// Option : secret code File
	apdu[7] = FDB_SECRET_CODE_FILE;
	apdu[8] = FDB_SECRET_CODE_FILE;

	// Body size
	apdu[9] = NULL_BYTE;
	apdu[10] = NB_SECRET_CODE * SECRET_CODE_SIZE;

	// ACL: These 3 AC will be locked futhermore

	// AC groupe 1 : free access
	apdu[11] = NO_SECRET_CODE_PROTECTION;
	apdu[12] = NULL_BYTE;

	// AC groupe 2 : free access
	apdu[13] = NO_SECRET_CODE_PROTECTION;
	apdu[14] = NULL_BYTE;

	// RFU
	apdu[15] = NO_SECRET_CODE_PROTECTION;
	apdu[16] = NULL_BYTE;

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 17, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC on Sending APDU Create Secret Code EF: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Creating Secret Code EF: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Create the Elementary File storing the user name
	DEBUG_MSG("Create the Elementary File storing the user name");
	///////////////////////////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = CREATE_FILE;
	apdu[2] = EF_TYPE;
	apdu[3] = NULL_BYTE;
	apdu[4] = CREATE_EF_DATA_SIZE;

	// Fill up the data buffer

	// SFI
	apdu[5] = PAM_DF_SFI;
	apdu[6] = USER_NAME_SFI;

	// Option : Transparent File
	apdu[7] = FDB_TRANSPARENT_FILE;
	apdu[8] = NULL_BYTE;

	// Body size is USER_MAX_SIZE
	apdu[9] = NULL_BYTE;
	apdu[10] = USER_NAME_MAX_SIZE;

	// ACL

	// AC groupe 1: Protected in updating by the administrator pin
	apdu[11] = PROTECTED_BY_ONE_PIN_CODE;
	apdu[12] = LOCAL_ADM_PIN_CODE;

	// AC groupe 2: Protected in writing by the administrator pin
	apdu[13] = PROTECTED_BY_ONE_PIN_CODE;
	apdu[14] = LOCAL_ADM_PIN_CODE;

	// AC groupe 3
	apdu[15] = NO_SECRET_CODE_PROTECTION;
	apdu[16] = NULL_BYTE;

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 17, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC on Sending APDU Create User Name EF: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Creating User Name EF: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Create the RSA keys file
	DEBUG_MSG("Create the RSA keys file");
	///////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = CREATE_FILE;
	apdu[2] = EF_TYPE;
	apdu[3] = NULL_BYTE;
	apdu[4] = CREATE_EF_DATA_SIZE;

	// Fill up the data buffer

	// SFI
	apdu[5] = PAM_DF_SFI;
	apdu[6] = RSA_KEY_SFI;

	// Option : RSA Keys
	apdu[7] = FDB_RSA_KEY_FILE;
	apdu[8] = NULL_BYTE;

	// Body size : Public part is stocked
	apdu[9] = NULL_BYTE;

	if (RSA_KEY_SIZE == 512)
		apdu[10] = CREATE_EF_RSA_KEY_512_SIZE;
	else
		apdu[10] = CREATE_EF_RSA_KEY_1024_SIZE;

	// ACL
	// The 2 next AC will be locked futhermore

	// AC groupe 1
	apdu[11] = NO_SECRET_CODE_PROTECTION;
	apdu[12] = NULL_BYTE;

	// AC groupe 2
	apdu[13] = NO_SECRET_CODE_PROTECTION;
	apdu[14] = NULL_BYTE;

	// AC groupe 3
	apdu[15] = PROTECTED_BY_ONE_PIN_CODE;	// RSA Key public part :
	apdu[16] = LOCAL_ADM_PIN_CODE;	// Protected in reading by the admin PIN

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 17, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC on Sending APDU Create RSA Keys EF: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Creating RSA Keys EF: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Create the CFS secret file
	DEBUG_MSG("Create the CFS secret file");
	/////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = CREATE_FILE;
	apdu[2] = EF_TYPE;
	apdu[3] = NULL_BYTE;
	apdu[4] = CREATE_EF_DATA_SIZE;

	// Fill up the data buffer

	// SFI
	apdu[5] = PAM_DF_SFI;
	apdu[6] = CFS_SECRET_SFI;

	// Option : Transparent File
	apdu[7] = FDB_TRANSPARENT_FILE;
	apdu[8] = NULL_BYTE;

	// Body size is CFS_SECRET_SIZE
	apdu[9] = NULL_BYTE;
	apdu[10] = CFS_SECRET_SIZE;

	// These 2 NEXT AC will be locked futhermore

	// AC groupe 1
	apdu[11] = NO_SECRET_CODE_PROTECTION;
	apdu[12] = NULL_BYTE;

	// AC groupe 2
	apdu[13] = NO_SECRET_CODE_PROTECTION;
	apdu[14] = NULL_BYTE;

	// AC groupe 3
	apdu[15] = PROTECTED_BY_ONE_PIN_CODE;
	apdu[16] = LOCAL_USR_PIN_CODE;	// Protected in reading by the user PIN

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 17, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC on Sending APDU Create CFS Secret EF: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Creating CFS Secret EF: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Load the user PIN : code number 0
	DEBUG_MSG("Load the user PIN : code number 0");
	// By default, each digit egal 0, this code will be set by another function.
	// Unlockable by the administrator PUC: code number 1
	////////////////////////////////////////////////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = UPDATE_BINARY;
	apdu[2] = P1_SELECT_UPDATE_FROM_EF | SECRET_CODE_SFI;
	apdu[3] = P2_OFFSET_NULL;
	apdu[4] = SECRET_CODE_SIZE / 2 + 4;

	// Fill up the data buffer

	// Max number of presentation
	apdu[5] = USR_PIN_MAX_PRES;
	apdu[6] = NULL_BYTE;

	// Option : Nr of PUC...
	apdu[7] = 0x08 | ADM_PUC_CODE_NUMBER;	// Unlockable by administrator PUC

	// PIN Value
	for (i = 0; i < SECRET_CODE_SIZE / 2; i++)
		apdu[9 + i] = NULL_BYTE;

	// Checksum calculus
	apdu[8] =
		apdu[5] ^ apdu[6] ^ apdu[7] ^ apdu[9] ^ apdu[10] ^ apdu[11] ^ apdu[12]
		^ 0xFF;

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 13, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC on Sending APDU Load User PIN Code: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Loading User PIN Code: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Load the administrator PUC : code number 1
	DEBUG_MSG("Load the administrator PUC : code number 1");
	/////////////////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = UPDATE_BINARY;
	apdu[2] = P1_SELECT_UPDATE_FROM_EF | SECRET_CODE_SFI;
	apdu[3] = 0x02;
	apdu[4] = SECRET_CODE_SIZE / 2 + 4;

	// Fill up the data buffer

	// Max number of presentation
	apdu[5] = ADM_PUC_MAX_PRES;
	apdu[6] = NULL_BYTE;

	// Option : 
	apdu[7] = 0x08 | ADM_PUC_CODE_NUMBER;
	// Unlockable by itself, so not unlockable

	// PUC Value
	for (i = 0; i < SECRET_CODE_SIZE / 2; i++)
	{
		int tmp;
		sscanf(admPuc + 2 * i, "%02x", &tmp);
		apdu[9 + i] = (unsigned char) tmp;
	}

	// Checksum calculus
	apdu[8] =
		apdu[5] ^ apdu[6] ^ apdu[7] ^ apdu[9] ^ apdu[10] ^ apdu[11] ^ apdu[12]
		^ 0xFF;

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 13, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC on Sending APDU Load Administrator PUC Code: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Loading Administrator PUC Code: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Load the administrator PIN : code number 2
	DEBUG_MSG("Load the administrator PIN : code number 2");
	// Unlockable by the administrator PUC : code number 1
	//////////////////////////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = UPDATE_BINARY;
	apdu[2] = P1_SELECT_UPDATE_FROM_EF | SECRET_CODE_SFI;
	apdu[3] = 0x04;
	apdu[4] = SECRET_CODE_SIZE / 2 + 4;

	// Fill up the data buffer

	// Max number of presentation
	apdu[5] = ADM_PIN_MAX_PRES;
	apdu[6] = NULL_BYTE;

	// Option : Nr of PUC
	apdu[7] = 0x08 | ADM_PUC_CODE_NUMBER;
	// Unlockable by the administrator PUC

	// PIN Value
	for (i = 0; i < SECRET_CODE_SIZE / 2; i++)
	{
		int tmp;
		sscanf(admPin + 2 * i, "%02x", &tmp);
		apdu[9 + i] = (unsigned char) tmp;
	}

	// Checksum calculus
	apdu[8] =
		apdu[5] ^ apdu[6] ^ apdu[7] ^ apdu[9] ^ apdu[10] ^ apdu[11] ^ apdu[12]
		^ 0xFF;

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 13, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC on Sending APDU Load Administrator PIN Code: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Loading Administrator PIN Code: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Generate the CFS secret
	DEBUG_MSG("Generate the CFS secret");
	//////////////////////////

	// Open the kernel random number generator.
	// The generated random numbers seem have goo cryptographics 
	// properties.
	fd = open("/dev/urandom", O_RDONLY);

	if (fd == -1)
	{
		printf("UNABLE to open the kernel random number generator: %s\n",
			strerror(errno));
		return (EXIT_FAILURE);
	}

	// Read bytes to fill up the apdu data
	for (i = 0; i < CFS_SECRET_SIZE; i++)
	{
		rCode = read(fd, &apdu[5 + i], 1);

		if (rCode == -1)
		{
			printf
				("UNABLE to read numbers from the kernel random number generator: %s\n",
				strerror(errno));
			pcsc_release();
			return (EXIT_FAILURE);
		}
	}

	close(fd);

	// Store the CFS Secret
	DEBUG_MSG("Store the CFS Secret");
	///////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = UPDATE_BINARY;
	apdu[2] = CFS_SECRET_SFI | P1_SELECT_UPDATE_FROM_EF;
	apdu[3] = P2_OFFSET_NULL;
	apdu[4] = CFS_SECRET_SIZE;

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, CFS_SECRET_SIZE + 5,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC on Sending APDU Update CFS Secret: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Updating CFS Secrete: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Initialize the RSA keys file
	DEBUG_MSG("Initialize the RSA keys file");
	rCode =
		initialize_and_generate_RSA_key(RSA_KEY_SIZE,
		PROTECTED_BY_ONE_PIN_CODE, LOCAL_USR_PIN_CODE, RSA_KEY_SFI);
	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// + CHANGE THIS +
	// Freeze some files
	////////////////////

	// Secret codes file : no access possible
	DEBUG_MSG("Secret codes file : no access possible");
	/////////////////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = FREEZE_AC;
	apdu[2] = P1_MODE_EF;
	apdu[3] = NULL_BYTE;
	apdu[4] = FREEZE_DATA_SIZE;

	// Fill up the data buffer

	// SFI
	apdu[5] = PAM_DF_SFI;
	apdu[6] = SECRET_CODE_SFI;

	// Lock the three AC's
	apdu[7] = LOCK_ACCESS_CONDITION;
	apdu[8] = LOCK_ACCESS_CONDITION;
	apdu[9] = LOCK_ACCESS_CONDITION;
	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, FREEZE_DATA_SIZE + 5,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC on Sending APDU lock Secret Code file: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Locking Secret Code file: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}





	// RSA private key : acces in reading protected by user PIN code.
	// Other access prohibited.
	DEBUG_MSG
		("RSA private key : acces in reading protected by user PIN code.");
	/////////////////////////////////////////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = FREEZE_AC;
	apdu[2] = P1_MODE_EF;
	apdu[3] = NULL_BYTE;
	apdu[4] = FREEZE_DATA_SIZE;

	// Fill up the data buffer

	// SFI
	apdu[5] = PAM_DF_SFI;
	apdu[6] = RSA_KEY_SFI;

	// Lock the three AC's
	apdu[7] = LOCK_ACCESS_CONDITION;
	apdu[8] = LOCK_ACCESS_CONDITION;
	apdu[9] = NULL_BYTE;
	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, FREEZE_DATA_SIZE + 5,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC on Sending APDU lock RSA private key file: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Locking RSA private key file: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// CFS secret: acces in reading protected by user pin code
	// Other access prohibited
	DEBUG_MSG("CFS secret: acces in reading protected by user pin code");
	//////////////////////////////////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = FREEZE_AC;
	apdu[2] = P1_MODE_EF;
	apdu[3] = NULL_BYTE;
	apdu[4] = FREEZE_DATA_SIZE;

	// Fill up the data buffer

	// SFI
	apdu[5] = PAM_DF_SFI;
	apdu[6] = CFS_SECRET_SFI;

	// Lock the three AC's
	apdu[7] = LOCK_ACCESS_CONDITION;
	apdu[8] = LOCK_ACCESS_CONDITION;
	apdu[9] = NULL_BYTE;
	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, FREEZE_DATA_SIZE + 5,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC on Sending APDU lock CFS Secret file: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Locking CFS Secret file: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	printf("PC/SC exiting formatcard: %s\n", pcsc_stringify_error(rCode));
	pcsc_release();
	return (EXIT_SUCCESS);

}


int generate_certification_part(char *passphrase, int passphraseLen)
{
	// This functions creates the administrator certification card.
	// This card will be used to sign the user public keys.
	// This card is also used to decrypt the CFS secret.
	// It just contains 2 RSA keys protected in reading by the 
	// administrator PIN (generated using the passphrase).
	///////////////////////////////////////////////////////////////

	int rCode, i;

	BYTE recv[BUF_SIZE];
	BYTE apdu[BUF_SIZE];
	DWORD recvlen;
	SCARD_IO_REQUEST pio_recv_pci;

	char admPin[SECRET_CODE_SIZE];
	char admPuc[SECRET_CODE_SIZE];
	char csn[CSN_SIZE];

	// Card connection
	DEBUG_INIT;
	DEBUG_MSG("Card connection");
	rCode = pcsc_init(0);

	if (rCode != SCARD_S_SUCCESS)
	{
		pcsc_release();
		return (rCode);
	}

	// Get the card serial number
	DEBUG_MSG("Get the card serial number");
	rCode = get_card_serial_number(csn);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Generate PIN and PUC codes
	DEBUG_MSG("Generate PIN and PUC codes");
	generate_code(admPin, passphrase, passphraseLen, csn, CERT_ADM_PIN_NR,
		ADM_PIN_SIZE);
	generate_code(admPuc, passphrase, passphraseLen, csn, CERT_ADM_PUC_NR,
		ADM_PUC_SIZE);

	// Create the CERT DF
	DEBUG_MSG("Create the CERT DF");
	rCode =
		create_df_with_secure_messaging(CERT_DF_NAME, CERT_DF_NAME_LEN,
		CERT_DF_SFI);
	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Select the CERT Dedicated File
	DEBUG_MSG("Select the CERT Dedicated File");
	rCode = select_DF(CERT_DF_SFI, 0);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Create the secret code file
	//////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = CREATE_FILE;
	apdu[2] = EF_TYPE;
	apdu[3] = NULL_BYTE;
	apdu[4] = CREATE_EF_DATA_SIZE;

	// Fill up the data buffer

	// SFI
	apdu[5] = CERT_DF_SFI;
	apdu[6] = CERT_SECRET_CODE_SFI;

	// Option : secret code File
	apdu[7] = FDB_SECRET_CODE_FILE;
	apdu[8] = NULL_BYTE;

	// Body size
	apdu[9] = NULL_BYTE;
	apdu[10] = CERT_NB_SECRET_CODE * SECRET_CODE_SIZE;

	// 24 bytes in order to store 2 codes
	// These 3 AC will be locked futhermore

	// AC groupe 1       
	apdu[11] = NO_SECRET_CODE_PROTECTION;
	apdu[12] = NULL_BYTE;

	// AC groupe 2
	apdu[13] = NO_SECRET_CODE_PROTECTION;
	apdu[14] = NULL_BYTE;

	// AC groupe 3
	apdu[15] = NO_SECRET_CODE_PROTECTION;
	apdu[16] = NULL_BYTE;

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 17, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Create Secret Code EF: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Creating Secret Code EF: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Create the first RSA keys file
	DEBUG_MSG("Create the first RSA keys file");
	/////////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = CREATE_FILE;
	apdu[2] = EF_TYPE;
	apdu[3] = NULL_BYTE;
	apdu[4] = CREATE_EF_DATA_SIZE;

	// Fill up the data buffer

	// SFI
	apdu[5] = CERT_DF_SFI;
	apdu[6] = CERT_RSA_KEY_SFI;

	// Option : RSA Keys
	apdu[7] = FDB_RSA_KEY_FILE;
	apdu[8] = NULL_BYTE;

	// Body size : Public part is stocked
	apdu[9] = NULL_BYTE;

	if (CERT_RSA_KEY_SIZE == 512)
		apdu[10] = CREATE_EF_RSA_KEY_512_SIZE;
	else
		apdu[10] = CREATE_EF_RSA_KEY_1024_SIZE;

	// The 2 next AC will be locked futhermore

	// AC groupe 1
	apdu[11] = NO_SECRET_CODE_PROTECTION;
	apdu[12] = NULL_BYTE;

	// AC groupe 2
	apdu[13] = NO_SECRET_CODE_PROTECTION;
	apdu[14] = NULL_BYTE;

	// AC groupe 3
	apdu[15] = PROTECTED_BY_ONE_PIN_CODE;	// RSA Key public part :
	apdu[16] = LOCAL_CERT_ADM_PIN_CODE;	// Protected in reading by the adm pin

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 17, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Create RSA Keys EF: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Creating RSA Keys EF: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Create the second RSA keys file
	DEBUG_MSG("Create the second RSA keys file");
	//////////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = CREATE_FILE;
	apdu[2] = EF_TYPE;
	apdu[3] = NULL_BYTE;
	apdu[4] = CREATE_EF_DATA_SIZE;

	// Fill up the data buffer

	// SFI
	apdu[5] = CERT_DF_SFI;
	apdu[6] = REST_RSA_KEY_SFI;

	// Option : RSA Keys
	apdu[7] = FDB_RSA_KEY_FILE;
	apdu[8] = NULL_BYTE;

	// Body size : Public part is stocked
	apdu[9] = NULL_BYTE;

	if (REST_RSA_KEY_SIZE == 512)
		apdu[10] = CREATE_EF_RSA_KEY_512_SIZE;
	else
		apdu[10] = CREATE_EF_RSA_KEY_1024_SIZE;

	// The 2 next AC will be locked futhermore

	// AC groupe 1
	apdu[11] = NO_SECRET_CODE_PROTECTION;
	apdu[12] = NULL_BYTE;

	// AC groupe 2
	apdu[13] = NO_SECRET_CODE_PROTECTION;
	apdu[14] = NULL_BYTE;

	// AC groupe 3
	apdu[15] = PROTECTED_BY_ONE_PIN_CODE;	// RSA Key public part :
	apdu[16] = LOCAL_CERT_ADM_PIN_CODE;	// Protected in reading by the adm pin

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 17, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Create RSA Keys EF: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Creating RSA Keys EF: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Load the PIN codes
	/////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = UPDATE_BINARY;
	apdu[2] = P1_SELECT_UPDATE_FROM_EF | CERT_SECRET_CODE_SFI;
	apdu[3] = NULL_BYTE;
	apdu[4] = SECRET_CODE_SIZE / 2 + 4;

	// Load the administrator PUC : code number 0
	DEBUG_MSG("Load the administrator PUC : code number 0");
	/////////////////////////////////////////////

	// Max number of presentation
	apdu[5] = CERT_ADM_PUC_MAX_PRES;
	apdu[6] = NULL_BYTE;

	// Option : 
	apdu[7] = 0x08 | ADM_PUC_CODE_NUMBER;	// Unlockable by itself... so not unlockable

	// PUC Value
	for (i = 0; i < SECRET_CODE_SIZE / 2; i++)
	{
		int tmp;
		sscanf(admPuc + 2 * i, "%02x", &tmp);
		apdu[9 + i] = (unsigned char) tmp;
	}

	// Checksum calculus
	apdu[8] =
		apdu[5] ^ apdu[6] ^ apdu[7] ^ apdu[9] ^ apdu[10] ^ apdu[11] ^ apdu[12]
		^ 0xFF;

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 13, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf
			("PC/SC Error Sending APDU Loading Administrator PUC Code: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Loading Administrator PUC Code: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Load the administrator PIN : code number 1
	// Unlockable by the administrator PUC : code number 0
	DEBUG_MSG("Load the administrator PIN : code number 1");
	//////////////////////////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = UPDATE_BINARY;
	apdu[2] = P1_SELECT_UPDATE_FROM_EF | CERT_SECRET_CODE_SFI;
	apdu[3] = 0x02;
	apdu[4] = SECRET_CODE_SIZE / 2 + 4;

	// Max number of presentation
	apdu[5] = CERT_ADM_PIN_MAX_PRES;
	apdu[6] = NULL_BYTE;

	// Option : Nr of PUC...
	apdu[7] = 0x08 | ADM_PUC_CODE_NUMBER;	// Unlockable by the administrator PUC

	// PIN Value
	for (i = 0; i < SECRET_CODE_SIZE / 2; i++)
	{
		int tmp;
		sscanf(admPin + 2 * i, "%02x", &tmp);
		apdu[9 + i] = (unsigned char) tmp;
	}

	// Checksum calculus
	apdu[8] =
		apdu[5] ^ apdu[6] ^ apdu[7] ^ apdu[9] ^ apdu[10] ^ apdu[11] ^ apdu[12]
		^ 0xFF;

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 13, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf
			("PC/SC Error Sending APDU Loading Administrator PIN Code: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Loading Administrator PIN Code: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Initialize the Certification RSA keys file
	DEBUG_MSG("Initialize the Certification RSA keys file");
	rCode =
		initialize_and_generate_RSA_key(CERT_RSA_KEY_SIZE,
		PROTECTED_BY_ONE_PIN_CODE, LOCAL_CERT_ADM_PIN_CODE, CERT_RSA_KEY_SFI);
	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Initialize the Escrow RSA keys file
	DEBUG_MSG("Initialize the Escrow RSA keys file");
	rCode =
		initialize_and_generate_RSA_key(REST_RSA_KEY_SIZE,
		PROTECTED_BY_ONE_PIN_CODE, LOCAL_CERT_ADM_PIN_CODE, REST_RSA_KEY_SFI);
	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);



	DEBUG_MSG("Freeze some files");
	////////////////////

	// Secret codes file : no access possible
	/////////////////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = FREEZE_AC;
	apdu[2] = P1_MODE_EF;
	apdu[3] = NULL_BYTE;
	apdu[4] = FREEZE_DATA_SIZE;

	// Fill up the data buffer

	// SFI
	apdu[5] = CERT_DF_SFI;
	apdu[6] = CERT_SECRET_CODE_SFI;

	// Lock the three AC's
	apdu[7] = LOCK_ACCESS_CONDITION;
	apdu[8] = LOCK_ACCESS_CONDITION;
	apdu[9] = LOCK_ACCESS_CONDITION;
	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, FREEZE_DATA_SIZE + 5,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC on Sending APDU Freeze Secret Code File: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Freezing Secret Code File: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// RSA private key : acces in reading protected by administrator
	// PIN code. Other access prohibited.
	DEBUG_MSG
		("RSA private key : acces in reading protected by administrator");
	////////////////////////////////////////////////////////////////

	// First RSA key file

	apdu[0] = PROPRIETARY;
	apdu[1] = FREEZE_AC;
	apdu[2] = P1_MODE_EF;
	apdu[3] = NULL_BYTE;
	apdu[4] = FREEZE_DATA_SIZE;

	// Fill up the data buffer

	// SFI
	apdu[5] = CERT_DF_SFI;
	apdu[6] = CERT_RSA_KEY_SFI;

	// Lock the three AC's
	apdu[7] = LOCK_ACCESS_CONDITION;
	apdu[8] = LOCK_ACCESS_CONDITION;
	apdu[9] = NULL_BYTE;
	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, FREEZE_DATA_SIZE + 5,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC on Sending APDU Freeze RSA Private Key File: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Freezing RSA Private Key File: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	// Second RSA key file
	DEBUG_MSG("Second RSA key file");

	apdu[0] = PROPRIETARY;
	apdu[1] = FREEZE_AC;
	apdu[2] = P1_MODE_EF;
	apdu[3] = NULL_BYTE;
	apdu[4] = FREEZE_DATA_SIZE;
	// Fill up the data buffer

	// SFI
	apdu[5] = CERT_DF_SFI;
	apdu[6] = REST_RSA_KEY_SFI;

	// Lock the three AC's
	apdu[7] = LOCK_ACCESS_CONDITION;
	apdu[8] = LOCK_ACCESS_CONDITION;
	apdu[9] = NULL_BYTE;
	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, FREEZE_DATA_SIZE + 5,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC on Sending APDU Freeze RSA Private Key File: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Freezing RSA Private Key File: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	printf("PC/SC exiting generate_certification_part: %s\n",
		pcsc_stringify_error(rCode));
	pcsc_release();
	return (EXIT_SUCCESS);
}

int get_card_serial_number(char *csn)
{
	// This function retrieves the card serial number and stores
	// it in csn.
	////////////////////////////////////////////////////////////

	int rCode;
	BYTE recv[BUF_SIZE];
	BYTE apdu[5];
	DWORD recvlen;
	SCARD_IO_REQUEST pio_recv_pci;

	apdu[0] = PROPRIETARY;
	apdu[1] = GET_INFO;
	apdu[2] = P1_CARD_SERIAL_NUMBER;
	apdu[3] = P2_CARD_SERIAL_NUMBER;
	apdu[4] = CSN_SIZE;

	recvlen = 10;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5, &pio_recv_pci, recv,
		&recvlen);


	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU get CSN: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Getting Card Ser Number: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	strncpy(csn, recv, recvlen - 2);
	csn[recvlen - 2] = 0x00;
	return (rCode);
}


int create_df_with_secure_messaging(char *DF_name, int DF_nameLen,
	unsigned char DF_sfi)
{
	// This function create a Dedicated File using the secure messaging
	// mechanism.
	///////////////////////////////////////////////////////////////////

	int i, rCode;

	BYTE recv[BUF_SIZE];
	BYTE apdu[43];
	DWORD recvlen;
	SCARD_IO_REQUEST pio_recv_pci;

	int dataSize;
	DESContext ks1, ks2;
	unsigned char card_session_respons[12];
	unsigned char session_key[16], RN[8];
	unsigned char key1[8], key2[8], key_prime[8], key_second[8];
	unsigned char iv[8], iv1[8], iv2[8];
	unsigned char data_to_sign[40], data_signed[40];

	// Set the session key
	//////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = SELECT_FK;
	apdu[2] = P1_SYSTEM_KEY;
	apdu[3] = P2_SYSTEM_KEY;
	apdu[4] = LC_SELECT_FK;
	for (i = 0; i < 8; i++)
	{
		// Random :)) Number
		apdu[5 + i] = NULL_BYTE;
	};

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 13, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Sending APDU SelectFK: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Selecting File Key : %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Get Response 
	////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = GET_RESPONSE;
	apdu[2] = NULL_BYTE;
	apdu[3] = NULL_BYTE;
	apdu[4] = recv[1];

	recvlen = recv[recvlen - 1] + 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error sending APDU Get Response: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Getting Response: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	memcpy(card_session_respons, recv, recvlen - 2);
	card_session_respons[recvlen - 2] = 0x00;

	// Session key calculus
	///////////////////////

	memcpy(RN, card_session_respons + 4, 8);
	memcpy(key1, SYSTEM_KEY, 8);
	memcpy(key2, SYSTEM_KEY + 8, 8);

	memset(iv, NULL_BYTE, 8);
	memset(iv1, NULL_BYTE, 8);
	memset(iv2, NULL_BYTE, 8);

	des_set_key(key1, &ks1);
	des_set_key(key2, &ks2);

	des_3cbc_encrypt(&ks1, iv, &ks2, iv1, &ks1, iv2, key_prime, RN, 8);

	memset(iv, NULL_BYTE, 8);
	memset(iv1, NULL_BYTE, 8);
	memset(iv2, NULL_BYTE, 8);

	des_3cbc_encrypt(&ks2, iv, &ks1, iv1, &ks2, iv2, key_second, RN, 8);

	memcpy(session_key, key_prime, 8);
	memcpy(session_key + 8, key_second, 8);

	// Create the Dedicated File with secure messaging
	//////////////////////////////////////////////////

	apdu[0] = PROPRIETARY_SECURE;
	apdu[1] = CREATE_FILE;
	apdu[2] = DF_TYPE;
	apdu[3] = NULL_BYTE;
	apdu[4] = 15 + DF_nameLen;

	// Fill up the data buffer

	// SFI
	apdu[5] = DF_sfi;
	apdu[6] = SFI_NULL;

	// Option
	apdu[7] = FDB_FOR_DF;
	apdu[8] = NULL_BYTE;

	// Name lenght
	apdu[9] = NULL_BYTE;
	apdu[10] = DF_nameLen;

	// AC groupe 1 : free access
	apdu[11] = NO_SECRET_CODE_PROTECTION;
	apdu[12] = NULL_BYTE;

	// AC groupe 2 : free access
	apdu[13] = NO_SECRET_CODE_PROTECTION;
	apdu[14] = NULL_BYTE;

	// RFU
	apdu[15] = NO_SECRET_CODE_PROTECTION;
	apdu[16] = NULL_BYTE;

	// DF name
	memcpy(apdu + 17, DF_name, DF_nameLen);

	// Checksum calculus: perform the CRYCKS Mechanism
	//////////////////////////////////////////////////

	memset(iv, NULL_BYTE, 8);
	memset(iv1, NULL_BYTE, 8);
	memset(iv2, NULL_BYTE, 8);

	memcpy(key1, session_key, 8);
	memcpy(key2, session_key + 8, 8);

	for (i = 0; i < 17 + DF_nameLen; i++)
		data_to_sign[i] = apdu[i];

	// Padd the end of the last byte with 0

	dataSize = 5 + 12 + DF_nameLen;

	if (dataSize % 8 != 0)
	{
		for (i = dataSize; i < dataSize + (8 - dataSize % 8); i++)
			data_to_sign[i] = NULL_BYTE;

		dataSize = dataSize + (8 - dataSize % 8);
	}

	des_set_key(key1, &ks1);
	des_set_key(key2, &ks2);

	des_3cbc_encrypt(&ks1, iv, &ks2, iv1, &ks1, iv2,
		data_signed, data_to_sign, dataSize);

	// Add the the LSB of CRYCKS at the end of data
	// The CRYCKS is the last 3 bytes of the signed datas
	apdu[17 + DF_nameLen] = data_signed[dataSize - 3];
	apdu[18 + DF_nameLen] = data_signed[dataSize - 2];
	apdu[19 + DF_nameLen] = data_signed[dataSize - 1];


	recvlen = 5;
	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 20 + DF_nameLen,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU createDFwSM: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}


	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Creating DF with SM: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}


int select_DF(unsigned char DF_SFI, int quiet)
{
	// This function selects the required Dedicated File
	////////////////////////////////////////////////////

	int rCode;

	BYTE recv[BUF_SIZE];
	BYTE apdu[7];
	DWORD recvlen;
	SCARD_IO_REQUEST pio_recv_pci;

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = SELECT_FILE;
	apdu[2] = P1_SELECT_DF_WHEN_MF_SELECTED;
	apdu[3] = P2_NO_REPONSE_REQUIRED;
	apdu[4] = SFI_SIZE;

	// Fill up the data buffer  
	apdu[5] = DF_SFI;
	apdu[6] = SFI_NULL;


	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 7, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		if (!quiet)
			printf("PC/SC Error Sending APDU Select DF number %02x: %s\n",
				DF_SFI, pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		if (!quiet)
			printf("Error Select DF number %02x: %s\n", DF_SFI,
				ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}


int initialize_and_generate_RSA_key(int rsaKeySize,
	unsigned char protectionMode, unsigned char codeNumber,
	unsigned char rsaKeySfi)
{
	// This function Initializes an RSA key file and runs the
	// on-board key generation.
	/////////////////////////////////////////////////////////

	int rCode;

	BYTE recv[BUF_SIZE];
	BYTE apdu[12];
	DWORD recvlen;
	SCARD_IO_REQUEST pio_recv_pci;

	DEBUG_MSG(" Append record");
	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = APPEND_RECORD;
	apdu[2] = NULL_BYTE;
	apdu[3] = rsaKeySfi << 3;
	apdu[4] = 0x07;

	// Initialize the RSA keys file
	///////////////////////////////

	// Fill up the data buffer

	// Data
	apdu[5] = SYS_REC_TAG;

	if (rsaKeySize == 512)
		apdu[6] = SYS_REC_KLEN_512;
	else
		apdu[6] = SYS_REC_KLEN_1024;

	apdu[7] = protectionMode;
	apdu[8] = codeNumber;

	apdu[9] = NULL_BYTE;
	apdu[10] = SYS_REC_ALGO_RSA;

	// Checksum Calculus
	apdu[11] = apdu[6] ^ apdu[7] ^ apdu[8] ^ apdu[9] ^ apdu[10] ^ 0xA5;

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 12, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Append RSA Record: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Appending RSA Record: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Create the RSA keys file private part
	////////////////////////////////////////

	DEBUG_MSG(" Create private key file");

	apdu[0] = PROPRIETARY;
	apdu[1] = CREATE_PRIVATE_KEY_FILE;
	apdu[2] = rsaKeySfi;
	if (rsaKeySize == 512)
		apdu[3] = P2_PRIVATE_512_SIZE;
	else
		apdu[3] = P2_PRIVATE_1024_SIZE;
	apdu[4] = NULL_BYTE;

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Create RSA Private Part: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Creating RSA Private Part: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Generate the RSA keys (on-board generation)
	//////////////////////////////////////////////

	DEBUG_MSG(" Generate RSA key");

	apdu[0] = PROPRIETARY;
	apdu[1] = GENERATE_RSA_KEY;
	apdu[2] = P1_SAVE_PUBLIC_PART | rsaKeySfi;
	if (rsaKeySize == 512)
		apdu[3] = P2_KEY_TYPE_512;
	else
		apdu[3] = P2_KEY_TYPE_1024;
	apdu[4] = 130;

	recvlen = 132;
	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Generate RSA Keys: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Generating the RSA Keys: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	/*
	   // ONLY FOR GPK8000 !
	   // Wait for the end of the keys generation
	   //////////////////////////////////////////

	   apdu[0] = ISO_COMPATIBLE;
	   apdu[1] = GET_RESPONSE;
	   apdu[2] = NULL_BYTE;
	   apdu[3] = NULL_BYTE;
	   apdu[4] = RSA_KEY_SIZE_IN_BYTES + 1;

	   recvlen = RSA_KEY_SIZE_IN_BYTES + 3;

	   nbReaded = -1;

	   while (nbReaded != 0) 
	   {
	   //unsigned char cardRespons[RSA_KEY_SIZE_IN_BYTES + 2];
	   //
	   //nbReaded =
	   //   SendOutgoingAPDUBinary(&desc, ISO_COMPATIBLE, GET_RESPONSE,
	   //   NULL_BYTE, NULL_BYTE, RSA_KEY_SIZE_IN_BYTES + 1, cardRespons,
	   //   &sw1, &sw2);
	   //
	   for (i=0; i<5; i++){printf("%02X",apdu[i]);};
	   printf("\n");
	   nbReaded = SCardTransmit(card, SCARD_PCI_T0, apdu, 5, &pio_recv_pci, recv, &recvlen);
	   sleep(1);
	   }
	 */

#ifdef BUGGY_GPK_KEY_GENERATION
	DEBUG_MSG(" buggy GPK arbitrary wait time");
#ifdef DEBUG_LEVEL_INFO
	{
		int i;

		for (i=BUGGY_GPK_KEY_GENERATION_TIME; i; i--)
		{
			printf("\r%d ", i);
			fflush(stdout);
			sleep(1);
		}
		printf("\n");
	}
#else
	sleep(BUGGY_GPK_KEY_GENERATION_TIME);
#endif
#endif

	return (EXIT_SUCCESS);
}

int install_certification_part(char *passphrase, int passphraseLen)
{
	// This function installs the administrator certification public key
	// and the administrator key able to restore the CFS secret.
	// These keys are stored on the certification card on the dedicated 
	// file on the computer.
	////////////////////////////////////////////////////////////////////

	int rCode, i;

	BYTE recv[BUF_SIZE];
	BYTE apdu[BUF_SIZE];
	DWORD recvlen;
	SCARD_IO_REQUEST pio_recv_pci;

	FILE *fd_pam, *fd_cfs;
	char csn[CSN_SIZE];
	char admPin[SECRET_CODE_SIZE];
	unsigned char restPublicKey[REST_RSA_KEY_SIZE_IN_BYTES];
	unsigned char restPublicKeyInvert[REST_RSA_KEY_SIZE_IN_BYTES + 1];
	unsigned char certPublicKey[CERT_RSA_KEY_SIZE_IN_BYTES];
	unsigned char certPublicKeyInvert[CERT_RSA_KEY_SIZE_IN_BYTES + 1];

	// Open the certification public key file
	fd_pam = fopen(ROOT_CERT_FILE_NAME, "w+");

	if (fd_pam == NULL)
	{
		printf("UNABLE to open the root certification key file (%s): %s\n",
			ROOT_CERT_FILE_NAME, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Open the restore CFS secret public key file
	fd_cfs = fopen(ROOT_REST_FILE_NAME, "w+");

	if (fd_cfs == NULL)
	{
		printf
			("UNABLE to open the root restore CFS secret key file (%s): %s\n",
			ROOT_REST_FILE_NAME, strerror(errno));
		return (EXIT_FAILURE);
	}

	// Card connection
	rCode = pcsc_init(0);

	if (rCode != SCARD_S_SUCCESS)
	{
		pcsc_release();
		return (rCode);
	}


	// Get the card serial number
	rCode = get_card_serial_number(csn);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Generate administrator PIN code
	generate_code(admPin, passphrase, passphraseLen, csn, CERT_ADM_PIN_NR,
		ADM_PIN_SIZE);

	// Select the CERT Dedicated File
	rCode = select_DF(CERT_DF_SFI, 0);

	if (rCode == EXIT_FAILURE)
		return (ERROR_CARD_NOT_FORMATED);

	// PIN checking in order to allow the reading of the RSA public key File
	////////////////////////////////////////////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = VERIFY;
	apdu[2] = P1_VERIFY;
	apdu[3] = CERT_ADM_PIN_CODE_NUMBER;
	apdu[4] = SECRET_CODE_SIZE;

	// Copy the administrator PIN in the data buffer
	for (i = 0; i < ADM_PIN_SIZE; i++)
		apdu[5 + i] = admPin[i] - '0';

	// Padd the end of the PIN with 0
	for (i = ADM_PIN_SIZE; i < SECRET_CODE_SIZE; i++)
		apdu[5 + i] = '0';
	recvlen = 2;
	rCode = SCardTransmit(card, SCARD_PCI_T0, apdu, 5 + SECRET_CODE_SIZE,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Verify Administrator PIN Code: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Verifying Administrator PIN Code: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Read the certification key public part record
	////////////////////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = READ_RECORD;
	apdu[2] = P1_PUBLIC_PART_RECORD_NR;
	apdu[3] = P2_CERT_PRIVATE_KEY_SFI;
	apdu[4] = RSA_KEY_SIZE_IN_BYTES + 1;
	recvlen = RSA_KEY_SIZE_IN_BYTES + 3;
	rCode = SCardTransmit(card, SCARD_PCI_T0, apdu, 5, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Read Record Certification Public Key: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Reading Record Certification Public Key: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	memcpy(certPublicKeyInvert, recv, RSA_KEY_SIZE_IN_BYTES + 1);
	DEBUG_XXD("admin key (gpk format) ", certPublicKeyInvert,
		RSA_KEY_SIZE_IN_BYTES + 1);

	// Read the restore key public part record
	//////////////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = READ_RECORD;
	apdu[2] = P1_PUBLIC_PART_RECORD_NR;
	apdu[3] = P2_REST_PRIVATE_KEY_SFI;
	apdu[4] = RSA_KEY_SIZE_IN_BYTES + 1;
	recvlen = RSA_KEY_SIZE_IN_BYTES + 3;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf
			("PC/SC Error Sending APDU Read Record Restore Public Key: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Reading Record Restore Public Key: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	memcpy(restPublicKeyInvert, recv, RSA_KEY_SIZE_IN_BYTES + 1);
	DEBUG_XXD("CFS key (gpk format) ", restPublicKeyInvert,
		RSA_KEY_SIZE_IN_BYTES + 1);

	// Disconnection
	pcsc_release();
	DEBUG_INFO2("PC/SC exiting install_certification_part: %s",
		pcsc_stringify_error(rCode));

	// Store the public key in the root certification key file
	//////////////////////////////////////////////////////////

	// Invert the key sent by the card and suppress the first byte
	for (i = 0; i < CERT_RSA_KEY_SIZE_IN_BYTES; i++)
		certPublicKey[i] = certPublicKeyInvert[CERT_RSA_KEY_SIZE_IN_BYTES - i];

	DEBUG_XXD("admin key (normal format) ", certPublicKey,
		RSA_KEY_SIZE_IN_BYTES);

	// Write the key in the file under the text format
	for (i = 0; i < CERT_RSA_KEY_SIZE_IN_BYTES; i++)
		if (fprintf(fd_pam, "%02X", (unsigned char) certPublicKey[i]) !=
			(2 * sizeof(unsigned char)))
		{
			printf("UNABLE to write the root certification key in file %s: %s\n",
				ROOT_CERT_FILE_NAME, strerror(errno));
			return (EXIT_FAILURE);
		}
	fprintf(fd_pam, "\n");
	fclose(fd_pam);

	// Store the other public key in the root restoration key file
	//////////////////////////////////////////////////////////////

	// Invert the key sent by the card and suppress the first byte
	for (i = 0; i < REST_RSA_KEY_SIZE_IN_BYTES; i++)
		restPublicKey[i] = restPublicKeyInvert[REST_RSA_KEY_SIZE_IN_BYTES - i];

	DEBUG_XXD("CFS key (normal format) ", restPublicKey, RSA_KEY_SIZE_IN_BYTES);

	// Write the key in the file under the text format
	for (i = 0; i < REST_RSA_KEY_SIZE_IN_BYTES; i++)
		if (fprintf(fd_cfs, "%02X", (unsigned char) restPublicKey[i]) !=
			(2 * sizeof(unsigned char)))
		{
			printf("UNABLE to write the root certification key in file %s: %s\n",
				ROOT_REST_FILE_NAME, strerror(errno));
			return (EXIT_FAILURE);
		}
	fprintf(fd_cfs, "\n");
	fclose(fd_cfs);

	return (EXIT_SUCCESS);
}

int perso_card(char *passphrase, int passphraseLen, char *userName,
	char *publicKeyDest, unsigned char *publicKey)
{
	int rCode, i;

	BYTE recv[BUF_SIZE];
	BYTE apdu[BUF_SIZE];
	DWORD recvlen;
	SCARD_IO_REQUEST pio_recv_pci;

	char csn[CSN_SIZE];
	char admPin[SECRET_CODE_SIZE];
	unsigned char publicKeyInvert[RSA_KEY_SIZE_IN_BYTES + 1];

	// Card connection
	rCode = pcsc_init(0);

	if (rCode != SCARD_S_SUCCESS)
	{
		pcsc_release();
		return (rCode);
	}

	// Get the card serial number
	rCode = get_card_serial_number(csn);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Generate administrator PIN code
	generate_code(admPin, passphrase, passphraseLen, csn, ADM_PIN_NR,
		ADM_PIN_SIZE);

	// Select the PAM Dedicated File
	rCode = select_DF(PAM_DF_SFI, 0);

	if (rCode == EXIT_FAILURE)
		return (ERROR_CARD_NOT_FORMATED);

	// PIN checking in order to allow the reading of the RSA public key File
	////////////////////////////////////////////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = VERIFY;
	apdu[2] = P1_VERIFY;
	apdu[3] = ADM_PIN_CODE_NUMBER;
	apdu[4] = SECRET_CODE_SIZE;

	for (i = 0; i < ADM_PIN_SIZE; i++)
		apdu[5 + i] = admPin[i] - '0';

	// Padd with 0

	for (i = ADM_PIN_SIZE; i < SECRET_CODE_SIZE; i++)
		apdu[5 + i] = '0';
	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5 + SECRET_CODE_SIZE,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Verify Administrator PIN Code: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Verifying Administrator PIN Code: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Update the user name stored in the card
	//////////////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = UPDATE_BINARY;
	apdu[2] = USER_NAME_SFI | P1_SELECT_UPDATE_FROM_EF;
	apdu[3] = P2_OFFSET_NULL;
	apdu[4] = USER_NAME_MAX_SIZE;

	memcpy(&apdu[5], userName, USER_NAME_MAX_SIZE);

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5 + USER_NAME_MAX_SIZE,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Update User Name: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Updating User Name: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Get the public key by reading the corresponding record
	/////////////////////////////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = READ_RECORD;
	apdu[2] = P1_PUBLIC_PART_RECORD_NR;
	apdu[3] = P2_PRIVATE_KEY_SFI;
	apdu[4] = RSA_KEY_SIZE_IN_BYTES + 1;

	recvlen = RSA_KEY_SIZE_IN_BYTES + 3;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Read Record RSA Public Key: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Reading Record RSA Public Key: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Disconnection
	pcsc_release();

	memcpy(publicKeyInvert, recv, RSA_KEY_SIZE_IN_BYTES + 1);

	// Invert the key and suppress the first byte

	for (i = 0; i < RSA_KEY_SIZE_IN_BYTES; i++)
		publicKey[i] = publicKeyInvert[RSA_KEY_SIZE_IN_BYTES - i];

	return (EXIT_SUCCESS);
}


int sign_user_public_key(unsigned char *publicKey, char *userName,
	char *passphrase, unsigned char *signature, int twoReaders)
{
	// This function is used to sign the user public key with the administrator
	// certification card. The signature is stored in signature.
	// The two readers flags indicates that the certification card is in the 
	// extra reader.
	///////////////////////////////////////////////////////////////////////////

	int rCode, i;

	BYTE recv[BUF_SIZE];
	BYTE apdu[BUF_SIZE];
	DWORD recvlen;
	SCARD_IO_REQUEST pio_recv_pci;

	char csn[CSN_SIZE];
	struct MD5Context md5Context;
	char admPin[SECRET_CODE_SIZE];
	unsigned char digest[DIGEST_SIZE];
	char publicKeyInTxt[RSA_KEY_SIZE_IN_TXT];
	char data[RSA_KEY_SIZE_IN_TXT + USER_NAME_MAX_SIZE];
	unsigned char dataSigned[HASH_PADDING_SIZE];

	// Card connection
	DEBUG_INIT;
	DEBUG_MSG("Card connection");
	if (twoReaders)
		rCode = pcsc_init(1);
	// The second reader must be in DEVICE_BIS  
	else
		rCode = pcsc_init(0);

	if (rCode != SCARD_S_SUCCESS)
	{
		pcsc_release();
		return (rCode);
	}

	// Get the card serial number
	DEBUG_MSG("Get the card serial number");
	rCode = get_card_serial_number(csn);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Generate administrator PIN code
	DEBUG_MSG("Generate administrator PIN code");
	generate_code(admPin, passphrase, strlen(passphrase), csn, CERT_ADM_PIN_NR,
		ADM_PIN_SIZE);

	// Select CERT Dedicated File
	DEBUG_MSG("Select CERT Dedicated File");
	rCode = select_DF(CERT_DF_SFI, 0);

	if (rCode == EXIT_FAILURE)
		return (ERROR_CARD_NOT_FORMATED);

	// Verify the administrator PIN code to gain
	// usage of the private key
	DEBUG_MSG("Verify the administrator PIN code");
	////////////////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = VERIFY;
	apdu[2] = P1_VERIFY;
	apdu[3] = CERT_ADM_PIN_CODE_NUMBER;
	apdu[4] = SECRET_CODE_SIZE;

	// Copy the PIN in the data buffer
	for (i = 0; i < ADM_PIN_SIZE; i++)
		apdu[5 + i] = admPin[i] - '0';

	// Padd the end oif the PIN with 0
	for (i = ADM_PIN_SIZE; i < SECRET_CODE_SIZE; i++)
		apdu[5 + i] = '0';
	recvlen = 2;
	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5 + SECRET_CODE_SIZE,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Verify Administrator PIN Code: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Verifying Administrator PIN Code: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Select the RSA private key file
	DEBUG_MSG("Select the RSA private key file");
	//////////////////////////////////

	apdu[0] = ISO_COMPATIBLE;
	apdu[1] = SELECT_FILE;
	apdu[2] = P1_SELECT_EF_UNDER_CURRENT_DF;
	apdu[3] = P2_NO_REPONSE_REQUIRED;
	apdu[4] = SFI_SIZE;

	// Fill up the data buffer
	apdu[5] = CERT_DF_SFI;
	apdu[6] = CERT_RSA_KEY_SFI;

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 7, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Select RSA Private Key File: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Selecting RSA Private Key File: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Select Crypto Context : Sign MD5 / RSA
	DEBUG_MSG("Select Crypto Context : Sign MD5 / RSA");
	/////////////////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = SELECT_CRYPTO_CONTEXT;
	apdu[2] = CERT_RSA_KEY_SFI;
	apdu[3] = P2_MODE_RSA_MD5_SIGN;
	apdu[4] = NULL_BYTE;

	// No data
	recvlen = 2;
	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Select Crypto Context: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Selecting Crypto Context: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Send the data to the card for signing.
	// Data to sign are the hashed of concatenation of the username 
	// and the user public key.
	DEBUG_MSG("Send the data to the card for signing.");
	///////////////////////////////////////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = INIT_HASHED_DATA;
	apdu[2] = NULL_BYTE;
	apdu[3] = NULL_BYTE;
	apdu[4] = DIGEST_SIZE;

	// Convert the user public key from byte format to text format
	for (i = 0; i < RSA_KEY_SIZE_IN_BYTES; i++)
		sprintf(publicKeyInTxt + 2 * i, "%02x", publicKey[i]);

	// Prepare the data; first part:username, second part: public key in text
	// pad to 0
	memset(data, 0, USER_NAME_MAX_SIZE);
	strncpy(data, userName, USER_NAME_MAX_SIZE);
	memcpy(data + USER_NAME_MAX_SIZE, publicKeyInTxt, RSA_KEY_SIZE_IN_TXT);

	DEBUG_XXD("data to hash: ", data, RSA_KEY_SIZE_IN_TXT + USER_NAME_MAX_SIZE);

	// Hash the data (using MD5)
	MD5Init(&md5Context);
	MD5Update(&md5Context, data, RSA_KEY_SIZE_IN_TXT + USER_NAME_MAX_SIZE);
	MD5Final(digest, &md5Context);

	DEBUG_XXD("hash to sign: ", digest, DIGEST_SIZE);

	// Invert the digest  
	for (i = 0; i < DIGEST_SIZE; i++)
		apdu[5 + i] = digest[DIGEST_SIZE - i - 1];

	// Send the hashed to the card.
	//////////////////////////////
	recvlen = 2;
	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5 + DIGEST_SIZE,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Init Hashed Data: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Initing Hashed Data: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Tell the card to sign the hash data
	DEBUG_MSG("Tell the card to sign the hash data");
	//////////////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = PK_SIGN;
	apdu[2] = NULL_BYTE;
	apdu[3] = NULL_BYTE;
	apdu[4] = CERT_HASH_PADDING_SIZE;
	recvlen = CERT_HASH_PADDING_SIZE + 2;
	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Sign the Data: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Signing the Data: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	memcpy(dataSigned, recv, CERT_HASH_PADDING_SIZE);

	DEBUG_XXD("signed data: ", dataSigned, CERT_HASH_PADDING_SIZE);

	// Copy the signature in the returned variable and 
	// invert it at the same time.
	for (i = 0; i < HASH_PADDING_SIZE; i++)
		signature[HASH_PADDING_SIZE - 1 - i] = dataSigned[i];

	// Disconnection
	DEBUG_MSG("Disconnection");
	pcsc_release();

	return (EXIT_SUCCESS);
}

int pin_mngmt(char *passphrase, int passphraseLen, char *newPin)
{
	int rCode, i;

	BYTE recv[BUF_SIZE];
	BYTE apdu[BUF_SIZE];
	DWORD recvlen;
	SCARD_IO_REQUEST pio_recv_pci;

	char csn[CSN_SIZE];
	char admPuc[SECRET_CODE_SIZE];

	// Card connection

	rCode = pcsc_init(0);

	if (rCode != SCARD_S_SUCCESS)
	{
		pcsc_release();
		return (rCode);
	}


	// Get the card serial number
	rCode = get_card_serial_number(csn);

	if (rCode == EXIT_FAILURE)
		return (ERROR_CARD_NOT_FORMATED);

	// Administrator PUC code generation 
	generate_code(admPuc, passphrase, passphraseLen, csn, ADM_PUC_NR,
		ADM_PUC_SIZE);

	// Select the PAM Dedicated File
	rCode = select_DF(PAM_DF_SFI, 0);

	if (rCode == EXIT_FAILURE)
		return (EXIT_FAILURE);

	// Send the unlock instruction with the new PIN
	///////////////////////////////////////////////

	apdu[0] = PROPRIETARY;
	apdu[1] = SET_SECRET_CODE;
	apdu[2] = P2_UNLOCK_SECRET_CODE;
	apdu[3] = USR_PIN_CODE_NUMBER;
	apdu[4] = SECRET_CODE_SIZE;

	// PUC Value
	for (i = 0; i < ADM_PUC_SIZE / 2; i++)
	{
		int tmp;
		sscanf(admPuc + 2 * i, "%02x", &tmp);
		apdu[5 + i] = (unsigned char) tmp;
	}

	// Padd the end of the PIN with \0
	for (i = ADM_PUC_SIZE / 2; i < SECRET_CODE_SIZE / 2; i++)
		apdu[5 + i] = 0x00;

	// New PIN Value
	for (i = 0; i < USR_PIN_SIZE / 2; i++)
	{
		int tmp;
		sscanf(newPin + 2 * i, "%02x", &tmp);
		apdu[5 + i + SECRET_CODE_SIZE / 2] = (unsigned char) tmp;
	}

	// Padd the end of the PIN with \0
	for (i = USR_PIN_SIZE / 2; i < SECRET_CODE_SIZE / 2; i++)
		apdu[5 + i + SECRET_CODE_SIZE / 2] = 0x00;
	recvlen = 2;
	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5 + SECRET_CODE_SIZE,
		&pio_recv_pci, recv, &recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Unlock User PIN Code: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error Unlocking User PIN Code: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	// Disconnection
	pcsc_release();

	return (EXIT_SUCCESS);
}


// ------------------------------->>>>NEW<<<<-----------------------------------

char *ISO7816_ErrorStr(unsigned char sw1, unsigned char sw2)
{
	static char iso7816_error[ISO7816_ERROR_TEXT_SIZE];
	char *ptr = iso7816_error;

	/* status words */
#ifdef WIN32

	_snprintf(ptr, ISO7816_ERROR_TEXT_SIZE - 1, "%02X%02X: ", sw1, sw2);

#else

	snprintf(ptr, ISO7816_ERROR_TEXT_SIZE - 1, "%02X%02X: ", sw1, sw2);

#endif


	/* go to the end of the string */
	ptr += strlen(ptr);

	if (sw1 == 0x90 && sw2 == 0x00)
		strncat(ptr, "Normal processing",
			ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1);


	if (sw1 == 0x61)
#ifdef WIN32
		_snprintf(ptr, ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1,
			"%d bytes of response still available.", sw2);
#else
		snprintf(ptr, ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1,
			"%d bytes of response still available.", sw2);
#endif

	if (sw1 == 0x62)
	{
		strncat(ptr, "State of non-volatile memory unchanged. ",
			ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1);

		if (sw2 == 0x00)
			strncat(ptr, "No informatin given.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x81)
			strncat(ptr, "Part of returned data my be corrupted.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x82)
			strncat(ptr,
				"End of file/record reached before reading Le bytes.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x83)
			strncat(ptr, "Selected file invalidated.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x84)
			strncat(ptr, "FCI not formated according to 5.1.5.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));
	}

	if (sw1 == 0x63)
	{
		strncat(ptr, "State of non-volatile memory changed. ",
			ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1);

		/* increment pointer for the snprintf() in the last case */
		ptr += strlen(ptr);

		if (sw2 == 0x00)
			strncat(ptr, "No informatin given.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x81)
			strncat(ptr, "File filled up by the last write.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if ((sw2 & 0xF0) == 0xC0)
#ifdef WIN32
			_snprintf(ptr,
				ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1,
				"Counter %d.", sw2 & 0x0F);
#else
			snprintf(ptr,
				ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1,
				"Counter %d.", sw2 & 0x0F);
#endif
	}

	if (sw1 == 0x64 && sw2 == 0x00)
		strncat(ptr,
			"State of non-volatile memory unchanged.",
			ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1);

	if (sw1 == 0x65)
	{
		strncat(ptr,
			"State of non-volatile memory changed. ",
			ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1);

		if (sw2 == 0x81)
			strncat(ptr, "Memory failure.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));
	}

	if (sw1 == 0x66)
		strncat(ptr,
			"Reserved for security-related issues.",
			ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1);

	if (sw1 == 0x67 && sw2 == 0x00)
		strncat(ptr, "Wrong length.",
			ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1);

	if (sw1 == 0x68 && sw2 == 0x00)
	{
		strncat(ptr, "Functions in CLA not supported. ",
			ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1);

		if (sw2 == 0x81)
			strncat(ptr, "Logical channel not supported.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x82)
			strncat(ptr, "Secure messaging not supported.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));
	}

	if (sw1 == 0x69)
	{
		strncat(ptr, "Command not allowed. ",
			ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1);

		if (sw2 == 0x81)
			strncat(ptr, "Command incompatible with file structure.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x82)
			strncat(ptr, "Security status not satisfied.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x83)
			strncat(ptr, "Authentication PIN locked.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x84)
			strncat(ptr, "Referenced data invalidated.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x85)
			strncat(ptr, "Conditions of use not satisfied.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x86)
			strncat(ptr, "Command not allowed (no current EF).",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x87)
			strncat(ptr, "Expected SM data objects missing.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x88)
			strncat(ptr, "SM dta objects incorrect.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));
	}

	if (sw1 == 0x6A)
	{
		strncat(ptr, "Wrong parameter(s) P1-P2. ",
			ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1);

		if (sw2 == 0x80)
			strncat(ptr, "Incorrect parameters in the data field.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x81)
			strncat(ptr, "Function not supported.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x82)
			strncat(ptr, "File not found.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x83)
			strncat(ptr, "Record not found.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x84)
			strncat(ptr, "Not enough memory space in the file.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x85)
			strncat(ptr, "Lc inconsistent with TLV structure.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x86)
			strncat(ptr, "Incorrect parameters P1-P2.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x87)
			strncat(ptr, "Lc inconsistent with P1-P2.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));

		if (sw2 == 0x88)
			strncat(ptr, "Referenced data not found.",
				ISO7816_ERROR_TEXT_SIZE - 1 - strlen(iso7816_error));
	}

	if (sw1 == 0x6B && sw2 == 0x00)
		strncat(ptr, "Wrong parameter(s) P1-P2.",
			ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1);

	if (sw1 == 0x6C)
#ifdef WIN32
		_snprintf(ptr, ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1,
			"Wrong length Le: should be %d", sw2);
#else
		snprintf(ptr, ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1,
			"Wrong length Le: should be %d", sw2);
#endif

	if (sw1 == 0x6D && sw2 == 0x00)
		strncat(ptr,
			"Instruction code not supported or invalid.",
			ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1);

	if (sw1 == 0x6E && sw2 == 0x00)
		strncat(ptr, "Class not supported.",
			ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1);

	if (sw1 == 0x6F && sw2 == 0x00)
		strncat(ptr, "No precise diagnosis.",
			ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1);

	/* default error message */
	if (strlen(iso7816_error) < 9)
		strncat(ptr, "Not defined by 7816.",
			ISO7816_ERROR_TEXT_SIZE - strlen(iso7816_error) - 1);

	return iso7816_error;
}								/* ISO-7816_ErrorStr */

int pcsc_init(int num)
{
	char *readers, *str;
	DWORD size, active;

	LONG rv;
	int i;

	//printf("In pcsc_init\n");
	/* establish context */
	rv = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &context);
	if (rv != SCARD_S_SUCCESS)
		return rv;

	/* change timeout */
	rv = SCardSetTimeout(context, 50);
	if (rv != SCARD_S_SUCCESS)
		return rv;

	/* get reader list size */
	rv = SCardListReaders(context, NULL, NULL, &size);
	if (rv != SCARD_S_SUCCESS)
		return rv;

	/* exit if there are no readers avaiable return */
	if (size == 0)
		return SCARD_E_READER_UNAVAILABLE;

	/* get all readers */
	readers = (char *) malloc(size);
	if (readers == NULL)
		return SCARD_E_READER_UNAVAILABLE;
	rv = SCardListReaders(context, NULL, readers, &size);
	if (rv != SCARD_S_SUCCESS)
	{
		free(readers);
		return rv;
	}

	/* name of reader number num */
	str = readers;
	for (i = 0; i < num; i++)
	{
		/* go to the next name */
		str += strlen(str) + 1;

		/* no more readers available? */
		if (strlen(str) == 0)
		{
			free(readers);
			return SCARD_E_READER_UNAVAILABLE;
		}
	}

	/* connect to selected reader */
	rv = SCardConnect(context, str, SCARD_SHARE_EXCLUSIVE, SCARD_PROTOCOL_T0,
		&card, &active);
	if (rv != SCARD_S_SUCCESS)
	{
		free(readers);
		return rv;
	}

	free(readers);
	return SCARD_S_SUCCESS;
}

void pcsc_release()
{
	LONG rv;
	//printf("In pcsc_release\n");
	/* disconnect from card, ignore result */
	rv = SCardDisconnect(card, SCARD_UNPOWER_CARD);

	/* release card context, ignore result */
	rv = SCardReleaseContext(context);
}


int erasecard()
{
	int rCode;

	BYTE recv[BUF_SIZE];
	BYTE apdu[5];
	DWORD recvlen;
	SCARD_IO_REQUEST pio_recv_pci;

	// Card connection
	DEBUG_INIT;
	DEBUG_MSG("Card connection");
	rCode = pcsc_init(0);

	if (rCode != SCARD_S_SUCCESS)
	{
		pcsc_release();
		return (rCode);
	}


	apdu[0] = PROPRIETARY_ERASE;
	apdu[1] = ERASE_CARD;
	apdu[2] = P1_ERASE_CARD;
	apdu[3] = P2_OFFSET_NULL;
	apdu[4] = NULL_BYTE;

	recvlen = 2;

	rCode =
		SCardTransmit(card, SCARD_PCI_T0, apdu, 5, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		printf("PC/SC Error Sending APDU Erase Card: %s\n",
			pcsc_stringify_error(rCode));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		printf("Error in ERASE_CARD: %s\n",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		pcsc_release();
		return (EXIT_FAILURE);
	}

	pcsc_release();
	return SCARD_S_SUCCESS;
}


int dircard()
{

	int rCode;
	BYTE apdu[BUF_SIZE];

/*		
	BYTE recv[BUF_SIZE];
    	DWORD recvlen;
    	SCARD_IO_REQUEST pio_recv_pci;
*/

	apdu[0] = 0x00;
	apdu[1] = 0xA4;
	apdu[2] = 0x01;
	apdu[3] = 0x00;
	apdu[4] = 0x02;

	rCode = pcsc_init(0);

	if (rCode != SCARD_S_SUCCESS)
	{
		pcsc_release();
		return (rCode);
	}

	printf("Not Implemented\n");


	pcsc_release();
	return (EXIT_FAILURE);
}

