/*
 *
 * This program is written by
 * Vincent Guerin
 * and Ludovic Rousseau <ludovic.rousseau@gemplus.com>
 *
 * $Id: apdu.h,v 1.5 2002/03/07 10:33:05 rousseau Exp $
 *
 */

/*
 * Copyright  2000 Gemplus
 * Av. du Pic de Bertagne - Parc d'Activits de Gmenos
 * BP 100 - 13881 Gmenos CEDEX
 * 
 * Everyone is allowed to redistribute and use this source  (source
 * code)  and binary (object code),  with or  without modification,
 * under some conditions:
 * 
 *  - Everyone  must  retain  and/or  reproduce the above copyright
 *    notice,  and the below  disclaimer of warranty and limitation
 *    of liability  for redistribution and use of these source code
 *    and object code.
 * 
 *  - Everyone  must  ask a  specific prior written permission from
 *    Gemplus to use the name of Gemplus.
 * 
 * DISCLAIMER OF WARRANTY
 * 
 * THIS CODE IS PROVIDED "AS IS",  WITHOUT ANY WARRANTY OF ANY KIND
 * (INCLUDING,  BUT  NOT  LIMITED  TO,  THE IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY  AND FITNESS FOR  A  PARTICULAR PURPOSE)  EITHER
 * EXPRESS OR IMPLIED.  GEMPLUS DOES NOT WARRANT THAT THE FUNCTIONS
 * CONTAINED  IN THIS SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR
 * THAT THE OPERATION OF IT WILL BE UNINTERRUPTED OR ERROR-FREE. NO
 * USE  OF  ANY  CODE  IS  AUTHORIZED  HEREUNDER EXCEPT UNDER  THIS
 * DISCLAIMER.
 * 
 * LIMITATION OF LIABILITY
 * 
 * GEMPLUS SHALL NOT BE LIABLE FOR INFRINGEMENTS OF  THIRD  PARTIES
 * RIGHTS. IN NO EVENTS, UNLESS REQUIRED BY APPLICABLE  LAW,  SHALL
 * GEMPLUS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES OF ANY CHARACTER  INCLUDING,
 * WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK STOPPAGE,
 * COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER DAMAGES OR
 * LOSSES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. ALSO,
 * GEMPLUS IS  UNDER NO  OBLIGATION TO MAINTAIN,  CORRECT,  UPDATE,
 * CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS SOFTWARE.
 */

#ifndef APDU_H
#define APDU_H

// This file defines the APDU (Application protocol Data
// Unit) codes used during the communication with the card.
// These codes are specific to the GPK 8000
///////////////////////////////////////////////////////////  

#define NULL_BYTE 0x00

// APDU Classe
//////////////

#define ISO_COMPATIBLE     0x00
#define ISO_SECURE         0x04
#define PROPRIETARY        0x80
#define PROPRIETARY_SECURE 0x84
#define PROPRIETARY_ERASE  0xDB

// APDU Instruction 
// and corresponding P1/P2 and possible parameters
//////////////////////////////////////////////////

// Erase Card
#define ERASE_CARD         0xDE
#define P1_ERASE_CARD      0x00

// Generate RSA Key
#define GENERATE_RSA_KEY    0xD2

#define P1_SAVE_PUBLIC_PART 0x80
#define P2_KEY_TYPE_512     0x00
#define P2_KEY_TYPE_1024    0x11

// Get Response
#define GET_RESPONSE 0xC0

// Create Private Key File
#define CREATE_PRIVATE_KEY_FILE 0x12

#define P2_PRIVATE_512_SIZE     0x2E
#define P2_PRIVATE_1024_SIZE    0x56

// Init RSA Key File arguments (Append record Arg)
#define SYS_REC_TAG        0x00

#define SYS_REC_KLEN_512   0x00
#define SYS_REC_KLEN_1024  0x11

#define SYS_REC_ALGO_RSA   0x00

#define PUBLIC_PART_LOCKED 0x30

// Append Record
#define APPEND_RECORD 0xE2

// Get Info
#define GET_INFO              0xC0

#define P1_CARD_SERIAL_NUMBER 0x02
#define P2_CARD_SERIAL_NUMBER 0xA0

// Update Binary
#define UPDATE_BINARY            0xD6

#define P1_SELECT_UPDATE_FROM_EF 0x80
#define P2_OFFSET_NULL           0x00

// Select File
#define SELECT_FILE                   0xA4

#define P1_SELECT_DF_WHEN_MF_SELECTED 0x01
#define P1_SELECT_EF_UNDER_CURRENT_DF 0x02
#define P2_NO_REPONSE_REQUIRED        0x0C

// Get Challenge
#define GET_CHALLENGE  0x84

#define SIZE_32_BYTES  0x20

// Read Binary
#define READ_BINARY            0xB0

#define P1_SELECT_READ_FROM_EF 0x80
#define P2_OFFSET_NULL         0x00

#define SFI_SIZE 0x02

// Freeze Access Condition
#define FREEZE_AC             0x16

#define P1_MODE_EF            0x02

#define FREEZE_DATA_SIZE      0x05
#define LOCK_ACCESS_CONDITION 0x40

// Verify
#define VERIFY    0x20

#define P1_VERIFY 0x00

// Select Crypto Contex
#define SELECT_CRYPTO_CONTEXT 0xA6

#define P2_MODE_RSA_MD5_SIGN  0x11
#define P2_MODE_RSA_UNWRAP    0x77

// Init Hashed Data
#define INIT_HASHED_DATA     0xEA

#define TAG_DIRECT_ADRESSING 0x55

// PK Sign
#define PK_SIGN 0x86

// PK Open envelope
#define PK_OPEN_ENVELOPE            0x1C

#define P2_RAW_ENVELOPE             0xFF

#define KEY_SIZE_1024               0x80

// Create File
#define CREATE_FILE                 0xE0

#define CREATE_EF_DATA_SIZE         0x0C

#define CREATE_EF_RSA_KEY_512_SIZE  0x4F
#define CREATE_EF_RSA_KEY_1024_SIZE 0x8F

// File Type
#define DF_TYPE 0x01
#define EF_TYPE 0x02

// FDB
#define FDB_SECRET_CODE_FILE 0x21
#define FDB_TRANSPARENT_FILE 0x01
#define FDB_RSA_KEY_FILE     0x2C
#define FDB_FOR_DF           0x38

// Select File key
#define SELECT_FK     0x28
#define P1_SYSTEM_KEY 0x00
#define P2_SYSTEM_KEY 0x01

#define LC_SELECT_FK 0x08

// Set Secret Code
#define SET_SECRET_CODE       0x24

#define P1_CHANGE_SECRET_CODE 0x00
#define P2_UNLOCK_SECRET_CODE 0x01

// Read Record
#define READ_RECORD              0xB2

#define P1_PUBLIC_PART_RECORD_NR 0x02

#define P2_PRIVATE_KEY_SFI       (0x04 | RSA_KEY_SFI<<3)

#define P2_CERT_PRIVATE_KEY_SFI  (0x04 | CERT_RSA_KEY_SFI<<3)
#define P2_REST_PRIVATE_KEY_SFI  (0x04 | REST_RSA_KEY_SFI<<3)

// The return codes SW1 & SW2
/////////////////////////////

#define SW1_PIN_ERROR             0x63
#define NO_SUCCESFULL_TRANSACTION 0x69
#define NB_TRIES_LEFT_BASE        0xC0
#define SUCCESS_NO_WAITING_DATA   0x90
#define SUCCESS_WITH_WAITING_DATA 0x61

#define Successfull_Command(sw1, sw2) ((sw1 == SUCCESS_NO_WAITING_DATA) \
		|| (sw1 ==  SUCCESS_WITH_WAITING_DATA))

// Acces conditions codes
/////////////////////////

#define ACCESS_NEVER_ALLOWED      0xC0
#define NO_SECRET_CODE_PROTECTION 0x00
#define PROTECTED_BY_ONE_PIN_CODE 0x40

#endif /* APDU_H */
