/*
 *
 * This program is written by
 * Vincent Guerin
 * and Ludovic Rousseau <ludovic.rousseau@gemplus.com>
 *
 * $Id: config.h,v 1.6 2002/03/12 13:39:52 rousseau Exp $
 *
 */

/*
 * Copyright  2000 Gemplus
 * Av. du Pic de Bertagne - Parc d'Activits de Gmenos
 * BP 100 - 13881 Gmenos CEDEX
 * 
 * Everyone is allowed to redistribute and use this source  (source
 * code)  and binary (object code),  with or  without modification,
 * under some conditions:
 * 
 *  - Everyone  must  retain  and/or  reproduce the above copyright
 *    notice,  and the below  disclaimer of warranty and limitation
 *    of liability  for redistribution and use of these source code
 *    and object code.
 * 
 *  - Everyone  must  ask a  specific prior written permission from
 *    Gemplus to use the name of Gemplus.
 * 
 * DISCLAIMER OF WARRANTY
 * 
 * THIS CODE IS PROVIDED "AS IS",  WITHOUT ANY WARRANTY OF ANY KIND
 * (INCLUDING,  BUT  NOT  LIMITED  TO,  THE IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY  AND FITNESS FOR  A  PARTICULAR PURPOSE)  EITHER
 * EXPRESS OR IMPLIED.  GEMPLUS DOES NOT WARRANT THAT THE FUNCTIONS
 * CONTAINED  IN THIS SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR
 * THAT THE OPERATION OF IT WILL BE UNINTERRUPTED OR ERROR-FREE. NO
 * USE  OF  ANY  CODE  IS  AUTHORIZED  HEREUNDER EXCEPT UNDER  THIS
 * DISCLAIMER.
 * 
 * LIMITATION OF LIABILITY
 * 
 * GEMPLUS SHALL NOT BE LIABLE FOR INFRINGEMENTS OF  THIRD  PARTIES
 * RIGHTS. IN NO EVENTS, UNLESS REQUIRED BY APPLICABLE  LAW,  SHALL
 * GEMPLUS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES OF ANY CHARACTER  INCLUDING,
 * WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK STOPPAGE,
 * COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER DAMAGES OR
 * LOSSES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. ALSO,
 * GEMPLUS IS  UNDER NO  OBLIGATION TO MAINTAIN,  CORRECT,  UPDATE,
 * CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS SOFTWARE.
 */

#ifndef CONFIG_H
#define CONFIG_H

// This file define some constants which could
// be modified in order to adapt the program to specific cases.
///////////////////////////////////////////////////////////////

// Concerning the authentication service
////////////////////////////////////////

// Pam authentication messages
#define INSERT_CARD_MESSAGE     "Please insert your card"
#define ENTER_PIN_MESSAGE       "Enter your PIN code:"

// Database default name
#define DEFAULT_DB_NAME    (SMARTSESSION_DIR "smartpamkey.dbm")

// Nis map name
#define NIS_KEY_MAP_NAME        "smartpamkey.byname"

// RSA Key Size (512 or 1024)
#define RSA_KEY_SIZE            1024

// Username size (16 usefull characters)
#define USER_NAME_MAX_SIZE      17

// User pin code size : 0 = no pin code and maximum size is 8
#define USR_PIN_SIZE            4

// Maximum number of user PIN code presentation, card is locked after
// This number must be between 1 and 9
#define USR_PIN_MAX_PRES        3

// Size of the administator passphrase
#define PASSPHRASE_MIN_LEN      8
#define PASSPHRASE_MAX_LEN      64


// Card and reader parameters
/////////////////////////////

// Location in the card of the PAM file directory
// Should be change in case of conflicts
#define PAM_DF_SFI      0x05


// Concerning the autolock service
//////////////////////////////////

// Location of the autolock binary, the autolock wrapper 
// and the tmp lock file
#define AUTOLOCK              "autolock"
#define AUTOLOCK_DIR          "/usr/sbin/"
#define PID_AUTOLOCK_FILE     "/var/run/autolock.pid"
#define PID_AUTOLOCK_FILE_TMP "/tmp/autolock.pid.tmp"

// Location of the autolock locker smartlocker binary
#define SMARTLOCKER_LOCATION     "/usr/sbin/smartlocker"

// Location of the autolock excluded user file
#define AUTOLOCK_EXCLUDED_FILE_NAME (SMARTSESSION_DIR "autolock_excluded")

// Delay of the autolock loop in seconds
// Short delay should monopolize two much the reader an 
// increase the risk of conflicts whith other application
// Long delay should be unsecure
#define AUTOLOCK_LOOP_DELAY      3


// Concerning the public keys certification part
////////////////////////////////////////////////

// Location of the root certification public key file
#define ROOT_CERT_FILE_NAME    (SMARTSESSION_DIR "pam_CA_key")

// Location of the root restore CFS secret public Key
#define ROOT_REST_FILE_NAME    (SMARTSESSION_DIR "CFS_deposite_key")

// Location in the card of the CERT file directory
// Should be change in case of conflicts
#define CERT_DF_SFI            0x06

// RSA certification key size (512 or 1024)
#define CERT_RSA_KEY_SIZE      1024

// RSA restore CFS secret key size (512 or 1024)
#define REST_RSA_KEY_SIZE      1024

// Concerning the CFS part
//////////////////////////

// Location of the CFS binaries
#define CFS_BIN_DIR "/usr/bin/"

// Name of the CFS binaries
#define CMKDIR  "cmkdir"
#define CATTACH "cattach"
#define CDETACH "cdetach"

// Location of the user cfs fstab file
#define CFS_FSTAB_NAME ".cfs_fstab"

// Location of the user cfs mtab database
#define CFS_MTAB_NAME  "/tmp/cfs_mtab"

// Location of the ciphered CFS secret
#define SAVED_SECRET_FILE ".saved_secret"

// Location of the restored CFS secret
#define RESTORED_SECRET_FILE "restored_secret"

// Max size of the user cfs mtab database location
#define CFS_FSTAB_LOCATION_MAX_SIZE 256

// This text will be inserted at the beginning of the cfs fstab 
// module on its creation. 
#define CFS_FSTAB_1_LINE "# This file gives the correspondance between the ciphered directory and"
#define CFS_FSTAB_2_LINE "# the clear mounting pount on /crypt/. Its format is:"
#define CFS_FSTAB_3_LINE "# ciphered_directory_name<tab>clear_name<tab>auto_attach_flag"
#define CFS_FSTAB_4_LINE "# "
#define CFS_FSTAB_5_LINE "# ciphered_directory_name must be an absolut path"
#define CFS_FSTAB_6_LINE "# auto_attach_flag is auto for pam auto attaching or noauto"

// The next parameters shouldn't be changed
// But who knows...
///////////////////////////////////////////

// The Gemplus default system key
#define SYSTEM_KEY               "TEST KEYTEST KEY"

// Max length a the database path
#define DB_NAME_MAX_LEN    100

// Extension of a dbm database file
#define DBM_FILE_EXTENSION       ".dbm"

// The smart session tools lock file
#define SST_LOCK_FILE            "/var/run/sst_lock"

// Gemplus public exponant... I think it will never change
#define RSA_PUBLIC_EXPOSANT     "010001"

// Time between two bells rings: 1s
#define TIME_BEFORE_BELL 1

// time between two beeps: 150 ms
#define TIME_BETWEEN_BELL 150

#endif /* CONFIG_H */
