/*
 *
 * This program is written by
 * Vincent Guerin
 * and Ludovic Rousseau <ludovic.rousseau@gemplus.com>
 *
 * $Id: global.h,v 1.4 2002/03/07 10:33:05 rousseau Exp $
 *
 */

/*
 * Copyright  2000 Gemplus
 * Av. du Pic de Bertagne - Parc d'Activits de Gmenos
 * BP 100 - 13881 Gmenos CEDEX
 * 
 * Everyone is allowed to redistribute and use this source  (source
 * code)  and binary (object code),  with or  without modification,
 * under some conditions:
 * 
 *  - Everyone  must  retain  and/or  reproduce the above copyright
 *    notice,  and the below  disclaimer of warranty and limitation
 *    of liability  for redistribution and use of these source code
 *    and object code.
 * 
 *  - Everyone  must  ask a  specific prior written permission from
 *    Gemplus to use the name of Gemplus.
 * 
 * DISCLAIMER OF WARRANTY
 * 
 * THIS CODE IS PROVIDED "AS IS",  WITHOUT ANY WARRANTY OF ANY KIND
 * (INCLUDING,  BUT  NOT  LIMITED  TO,  THE IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY  AND FITNESS FOR  A  PARTICULAR PURPOSE)  EITHER
 * EXPRESS OR IMPLIED.  GEMPLUS DOES NOT WARRANT THAT THE FUNCTIONS
 * CONTAINED  IN THIS SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR
 * THAT THE OPERATION OF IT WILL BE UNINTERRUPTED OR ERROR-FREE. NO
 * USE  OF  ANY  CODE  IS  AUTHORIZED  HEREUNDER EXCEPT UNDER  THIS
 * DISCLAIMER.
 * 
 * LIMITATION OF LIABILITY
 * 
 * GEMPLUS SHALL NOT BE LIABLE FOR INFRINGEMENTS OF  THIRD  PARTIES
 * RIGHTS. IN NO EVENTS, UNLESS REQUIRED BY APPLICABLE  LAW,  SHALL
 * GEMPLUS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES OF ANY CHARACTER  INCLUDING,
 * WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK STOPPAGE,
 * COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER DAMAGES OR
 * LOSSES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. ALSO,
 * GEMPLUS IS  UNDER NO  OBLIGATION TO MAINTAIN,  CORRECT,  UPDATE,
 * CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS SOFTWARE.
 */

#ifndef GLOBAL_H
#define GLOBAL_H

// This file define a lot of constant used by the functions
// It shouldn't be modified...
///////////////////////////////////////////////////////////

// Secret code specific defines
///////////////////////////////

// Size of a secret code in the card
#define SECRET_CODE_SIZE         8

// Size of the administator PIN and PUC code
#define ADM_PUC_SIZE             8
#define ADM_PIN_SIZE             8

// Maximum number of presentation of this codes
#define ADM_PUC_MAX_PRES         3
#define ADM_PIN_MAX_PRES         3

// For the CFS service
//////////////////////

// Size of the CFS secret
#define CFS_SECRET_SIZE        128

// Max lenght of a CFS idrectory name (including the path)
#define CFS_DIR_NAME_SIZE      512

// For the PAM CFS service
//////////////////////////

// The name of the data in the pam handle structure
#define PAM_PIN_DATA_NAME   "pin"

// Location of the CFS SC binaries
#define CFS_SC_BIN_DIR "/usr/bin/"

// Name of the CFS SC binaries
#define CATTACH_SC "cattach_SC"
#define CDETACH_SC "cdetach_SC"
#define CMKDIR_SC  "cmkdir_SC"

// Some size of buffers and other structures
////////////////////////////////////////////

// The size of a hash and padding text
#define HASH_PADDING_SIZE        RSA_KEY_SIZE_IN_BYTES

// The size of a hash and padding text written in hexadecimal
#define HASH_PADDING_SIZE_IN_TXT (HASH_PADDING_SIZE * 2)

// To have the RSA key size in byte
#define RSA_KEY_SIZE_IN_BYTES    (RSA_KEY_SIZE / 8)

// To have the RSA key size when it's write in hexadecimal
#define RSA_KEY_SIZE_IN_TXT      (RSA_KEY_SIZE_IN_BYTES * 2)

// The size of a MD5 digest
#define DIGEST_SIZE              16

// Maximum size of a database name
// It's a name given by the user
#define MAX_LEN_OF_DB_NAME FILENAME_MAX

// Maximum size of a autolock excluded user file line
#define MAX_LINE_LENGTH ( 2 * USER_NAME_MAX_SIZE )

// Size of the reference number
// This number is the base number used for the autentication
#define SIZEOF_REF_NUMBER        0x3E

// This of the temporary buffer
// This size have been calculated to be sufficient in all case
#define BUF_TMP_SIZE             256

// It's the size of the buffer used for store the data of the APDU
// defined in pcsclite.h
#define BUF_SIZE                 MAX_BUFFER_SIZE

// Size of the card serial number (Defined in the GPK8000 documentation)
#define CSN_SIZE                 8

// The command for the call of the autolock deamon
// by the pam_smartcard module
#define ADD_SESSION_CMD "add_session"
#define RM_SESSION_CMD  "rm_session"

// These constants are more specific to the card
////////////////////////////////////////////////

// The name of our dedicated file in the card, its length (7 chars max)
#define PAM_DF_NAME              "PAM DIR"
#define PAM_DF_NAME_LEN          strlen(PAM_DF_NAME)

// Number of the file present in the Pam directory
#define SFI_NULL        0x00	// Generic file, not a real file
#define SECRET_CODE_SFI 0x01	// Store the 3 secret codes
#define USER_NAME_SFI   0x02	// Store the user Name
#define RSA_KEY_SFI     0x03	// Store the RSA of the authentication key
#define CFS_SECRET_SFI  0x04	// Store the cfs secret

// Number of the 3 secret code in our secret code file
#define USR_PIN_CODE_NUMBER 0x00
#define ADM_PUC_CODE_NUMBER 0x01
#define ADM_PIN_CODE_NUMBER 0x02

// Secret code number with it's flag local
// Use for acces condition specification
#define LOCAL_USR_PIN_CODE (0x80 | USR_PIN_CODE_NUMBER<<4)
#define LOCAL_ADM_PUC_CODE (0x80 | ADM_PUC_CODE_NUMBER<<4)
#define LOCAL_ADM_PIN_CODE (0x80 | ADM_PIN_CODE_NUMBER<<4)

// Number of secret code in our secret code file
#define NB_SECRET_CODE 3

// Concerning the administrator certification card
//////////////////////////////////////////////////

// To have the RSA certification key size in byte
#define CERT_RSA_KEY_SIZE_IN_BYTES    (CERT_RSA_KEY_SIZE / 8)
#define CERT_RSA_KEY_SIZE_IN_TXT      (CERT_RSA_KEY_SIZE_IN_BYTES * 2)

#define REST_RSA_KEY_SIZE_IN_BYTES    (REST_RSA_KEY_SIZE / 8)
#define REST_RSA_KEY_SIZE_IN_TXT      (REST_RSA_KEY_SIZE_IN_BYTES * 2)

// Number of the 2 secret codes in the secret code file
#define CERT_ADM_PUC_CODE_NUMBER 0x00
#define CERT_ADM_PIN_CODE_NUMBER 0x01

// Secret code number with it's flag local
// Use for acces condition specification
#define LOCAL_CERT_ADM_PUC_CODE (0x80 | CERT_ADM_PUC_CODE_NUMBER<<4)
#define LOCAL_CERT_ADM_PIN_CODE (0x80 | CERT_ADM_PIN_CODE_NUMBER<<4)

// Number of secret code in the secret code file
#define CERT_NB_SECRET_CODE      2

// The name of the certification dedicated file in the card, its length (7 chars max)
#define CERT_DF_NAME              "CERTKEY"
#define CERT_DF_NAME_LEN          strlen(CERT_DF_NAME)

// Number of the file present in the Pam directory
#define CERT_SECRET_CODE_SFI 0x01	// Store the 2 secret codes
#define CERT_RSA_KEY_SFI     0x02	// Store the RSA authentication key
#define REST_RSA_KEY_SFI     0x03	// Store the RSA restauration key

// Maximun number of presentration of the administrator code
#define CERT_ADM_PUC_MAX_PRES         3
#define CERT_ADM_PIN_MAX_PRES         3

#define CERT_HASH_PADDING_SIZE        CERT_RSA_KEY_SIZE_IN_BYTES

#endif /* GLOBAL_H */
