/*
 *
 * This program is written by
 * Vincent Guerin
 * and Ludovic Rousseau <ludovic.rousseau@gemplus.com>
 *
 * Code changed to use PCSC instead of TLP calls
 * November 2001 - Domenico Taddei <domenico.taddei@proteam-group.it>
 *
 * $Id: autolock.c,v 1.12 2002/03/14 07:44:28 rousseau Exp $
 *
 */

/*
 * Copyright  2000 Gemplus
 * Av. du Pic de Bertagne - Parc d'Activits de Gmenos
 * BP 100 - 13881 Gmenos CEDEX
 * 
 * Everyone is allowed to redistribute and use this source  (source
 * code)  and binary (object code),  with or  without modification,
 * under some conditions:
 * 
 *  - Everyone  must  retain  and/or  reproduce the above copyright
 *    notice,  and the below  disclaimer of warranty and limitation
 *    of liability  for redistribution and use of these source code
 *    and object code.
 * 
 *  - Everyone  must  ask a  specific prior written permission from
 *    Gemplus to use the name of Gemplus.
 * 
 * DISCLAIMER OF WARRANTY
 * 
 * THIS CODE IS PROVIDED "AS IS",  WITHOUT ANY WARRANTY OF ANY KIND
 * (INCLUDING,  BUT  NOT  LIMITED  TO,  THE IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY  AND FITNESS FOR  A  PARTICULAR PURPOSE)  EITHER
 * EXPRESS OR IMPLIED.  GEMPLUS DOES NOT WARRANT THAT THE FUNCTIONS
 * CONTAINED  IN THIS SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR
 * THAT THE OPERATION OF IT WILL BE UNINTERRUPTED OR ERROR-FREE. NO
 * USE  OF  ANY  CODE  IS  AUTHORIZED  HEREUNDER EXCEPT UNDER  THIS
 * DISCLAIMER.
 * 
 * LIMITATION OF LIABILITY
 * 
 * GEMPLUS SHALL NOT BE LIABLE FOR INFRINGEMENTS OF  THIRD  PARTIES
 * RIGHTS. IN NO EVENTS, UNLESS REQUIRED BY APPLICABLE  LAW,  SHALL
 * GEMPLUS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES OF ANY CHARACTER  INCLUDING,
 * WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK STOPPAGE,
 * COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER DAMAGES OR
 * LOSSES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. ALSO,
 * GEMPLUS IS  UNDER NO  OBLIGATION TO MAINTAIN,  CORRECT,  UPDATE,
 * CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <string.h>
#include <pwd.h>
#include <fcntl.h>

#include "config.h"
#include "global.h"
#include "pam_smartcard_utils.h"
#include "smartcard_functions.h"
#include "apdu.h"

// The daemon autolock tests continously the presence of the user card
// in the reader. So, if there is no card or if the card have not
// the required card serial number (given is argument), it execute
// the screen locker smartlocker (under the name of the user given 
// in argument). After the end of the locker, it restarts.
//
// If it receives SIGUSR1, it stops.
// If it receives SIGUSR2, it restarts.
//////////////////////////////////////////////////////////////////////

// Internals functions
//////////////////////

// These one are used to handle the signals
///////////////////////////////////////////

// Handler for the suspend signal (SIG USR1)
void sig_usr1_handler(int sig);

// Handler for the resume signal (SIG USR2)
void sig_usr2_handler(int sig);

// Handler for the exit signal (SIG QUIT)
void sig_quit_handler(int sig);

// Internal function testing the presence of the good card in the reader
int is_good_card_in_reader(char *csnRequiredTxt);

// This function stops autolock and wait until the user pick up his card
int stopAutolock(void);

// Test loop, test the presence of the card in the reader
void autolock_loop(char *userName, char *csnRequiredTxt, int sessionPid);

// Adds a new session on tre list and executes the test loop
// if it's the first one
void add_session(char *userName, char *csnRequiredTxt, int sessionPid);

// Removes a registred session from the list and stops the
// daemon if it's the last one
void remove_session(int sessionPid);


// Global variables
///////////////////

// The suspended flag
int Suspended = 0;

// fd for /dev/console
int ConsoleFd = -1;


int main(int argc, char *argv[])
{
	char *userName;
	int sessionPid;
	char *csnRequiredTxt;

#if 0
	int i;

	printf("argc: %d\n", argc);
	for (i=0; i<argc; i++)
		printf("%d: %s\n", i, argv[i]);
#endif

	// Get the arguments 
	// and manage the command
	/////////////////////////

	switch (argc)
	{
		case 6:
			// It's the add session command
			userName = argv[2];
			csnRequiredTxt = argv[3];
			sessionPid = atoi(argv[4]);
			ConsoleFd = atoi(argv[5]);

			add_session(userName, csnRequiredTxt, sessionPid);
			break;

		case 3:
			// It's the remove session command
			sessionPid = atoi(argv[2]);

			remove_session(sessionPid);
			break;

		default:
			fprintf(stderr,
				"\nWrong number of arguments\n" \
				"Are you sure to want to run autolock?\n" \
				"This program should be exclusively used by pam_smartcard(5)\n");

			return (EXIT_FAILURE);
	}

	// We have to return something...
	return (EXIT_SUCCESS);
}

int is_good_card_in_reader(char *csnRequiredTxt)
{
	// Is the authenticated card always present in the reader?
	//////////////////////////////////////////////////////////

	int i, rCode;

	BYTE recv[BUF_SIZE];
	BYTE apdu[BUF_SIZE];
	DWORD recvlen;
	SCARD_IO_REQUEST pio_recv_pci;

	unsigned char csnReaded[CSN_SIZE];
	char csnReadedTxt[CSN_SIZE * 2 + 1];

	// Initialize the card serial number
	memset(csnReaded, '\0', CSN_SIZE);

	// Try to be connected to the card
	rCode = pcsc_init(0);

	if (rCode != SCARD_S_SUCCESS)
	{
		// Probably the card isn'nt in the reader
		pcsc_release();
		pam_log(AUTOLOCK_IDENT, LOG_INFO,
			"UNABLE to be connected to the card: locking the computer");
		return (EXIT_FAILURE);
	}

	// Get the card serial number
	apdu[0] = PROPRIETARY;
	apdu[1] = GET_INFO;
	apdu[2] = P1_CARD_SERIAL_NUMBER;
	apdu[3] = P2_CARD_SERIAL_NUMBER;
	apdu[4] = CSN_SIZE;
	recvlen = CSN_SIZE + 2;
	rCode = SCardTransmit(card, SCARD_PCI_T0, apdu, 5, &pio_recv_pci, recv,
		&recvlen);

	if (rCode != SCARD_S_SUCCESS)
	{
		pcsc_release();
		pam_log(AUTOLOCK_IDENT, LOG_INFO,
			"UNABLE to send get CSN APDU to the card (%s): locking the computer",
			pcsc_stringify_error(rCode));
		return (EXIT_FAILURE);
	}

	if (!Successfull_Command(recv[recvlen - 2], recv[recvlen - 1]))
	{
		pcsc_release();
		pam_log(AUTOLOCK_IDENT, LOG_INFO,
			"UNABLE to get the card serial number (%s): locking the computer",
			ISO7816_ErrorStr(recv[recvlen - 2], recv[recvlen - 1]));
		return (EXIT_FAILURE);
	}

	memcpy(csnReaded, recv, CSN_SIZE);
	// Card disconnection
	pcsc_release();

	// Compare the required card number and the readed card number
	// in order to be sure that card in the reader is the card
	// that has been authenticated
	//////////////////////////////////////////////////////////////

	// Convert the card serial number in text
	for (i = 0; i < CSN_SIZE; i++)
		sprintf(csnReadedTxt + 2 * i, "%02x", csnReaded[i]);
	csnReadedTxt[CSN_SIZE * 2] = '\0';

	if (strcmp(csnRequiredTxt, csnReadedTxt) == 0)
		return (EXIT_SUCCESS);
	else
		return (EXIT_FAILURE);
}

void sig_usr1_handler(int sig)
{
	// Card disconnection. It's perhaps necessary if
	// daemon is interrupted when it's connected with the card
	pcsc_release();

	pam_log(AUTOLOCK_IDENT, LOG_INFO, "Autolock daemon suspended\n");

	// Re-arm the handler
	signal(SIGUSR1, sig_usr1_handler);

	// Switch on the supended flag
	Suspended = 1;
}

void sig_usr2_handler(int sig)
{
	// Re-arm the handler
	signal(SIGUSR2, sig_usr2_handler);

	pam_log(AUTOLOCK_IDENT, LOG_INFO, "Autolock daemon resumed\n");

	// Switch off the supended flag
	Suspended = 0;
}

void sig_quit_handler(int sig)
{
	int rCode;
	SCARDCONTEXT hContext;
	DWORD dwReaders;
	LPSTR mszReaders;
	SCARD_READERSTATE_A rgReaderStates_t, rgReaderStates;

	pam_log(AUTOLOCK_IDENT, LOG_INFO, "Received quit(%d) signal", sig);

	// Card disconnection. It's perhaps necessary if
	// daemon is interrupted when it's connected with the card
	//pcsc_release();

	// Remove the autolock pid file
	rCode = truncate(PID_AUTOLOCK_FILE, 0);

	if (rCode == -1)
	{
		pam_log(AUTOLOCK_IDENT, LOG_ERR,
			"UNABLE to truncate the autolock pid file (%s): %s",
			PID_AUTOLOCK_FILE, strerror(errno));
		exit(EXIT_FAILURE);
	}

	// Now user can remove his card...
	// Inform the user by ringing a beep.
	if (ConsoleFd != -1)
		write(ConsoleFd, "\007", 1);
	else
		exit(EXIT_FAILURE);

	// This part of code is designed to warn the user that 
	// he forgets his card in the reader at the end of his session
	//////////////////////////////////////////////////////////////

	// Wait a little time and test the card removing
	sleep(TIME_BEFORE_BELL);

	/* establish context */
	rCode = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &hContext);
	if (rCode != SCARD_S_SUCCESS)
	{
		pam_log(AUTOLOCK_IDENT, LOG_ERR, "SCardEstablishContext: %s",
			pcsc_stringify_error(rCode));
		exit (EXIT_FAILURE);
	}

	rCode = SCardListReaders(hContext, NULL, NULL, &dwReaders);
	if (rCode != SCARD_S_SUCCESS)
	{
		pam_log(AUTOLOCK_IDENT, LOG_ERR, "SCardListReaders: %s",
			pcsc_stringify_error(rCode));
		exit (EXIT_FAILURE);
	}

	mszReaders = malloc(sizeof(char)*dwReaders);
	if (mszReaders == NULL)
	{
		pam_log(AUTOLOCK_IDENT, LOG_ERR, "malloc: not enough memory");
		exit (EXIT_FAILURE);
	}

	rCode = SCardListReaders(hContext, NULL, mszReaders, &dwReaders);
	if (rCode != SCARD_S_SUCCESS)
	{
		pam_log(AUTOLOCK_IDENT, LOG_ERR, "SCardListReaders: %s",
			pcsc_stringify_error(rCode));
		exit (EXIT_FAILURE);
	}

	if (mszReaders[0] == '\0')
	{
		pam_log(AUTOLOCK_IDENT, LOG_ERR, "No reader found");
		exit (EXIT_FAILURE);
	}

	// While the card isn't removed, ring the bell
	//////////////////////////////////////////////

	pam_log(AUTOLOCK_IDENT, LOG_INFO, "autolock beep");

	/* set the initial states */
	rgReaderStates_t.szReader = mszReaders;
	rgReaderStates_t.dwCurrentState = SCARD_STATE_EMPTY;

	while (1)
	{
		rgReaderStates = rgReaderStates_t;
		
		SCardGetStatusChange(hContext, TIME_BETWEEN_BELL, &rgReaderStates, 1);
		
		// beep
		write(ConsoleFd, "\007", 1);

		if (rgReaderStates.dwCurrentState != rgReaderStates.dwEventState)
		{
			rgReaderStates.dwCurrentState = rgReaderStates.dwEventState;
			rgReaderStates_t = rgReaderStates;

			if (rgReaderStates.dwEventState & SCARD_STATE_EMPTY)
				goto out;
		}
	}

out:
	close(ConsoleFd);

	pam_log(AUTOLOCK_IDENT, LOG_ALERT, "Autolock daemon stopped");

	exit(EXIT_SUCCESS);
}

void add_session(char *userName, char *csnRequiredTxt, int sessionPid)
{
	// Check if the autolog pid file exist? 
	//   No  --> It's the first session, switch on autolock.
	//   Yes --> Autolock is already running. 
	//////////////////////////////////////////////////////////

	int pid;
	int rCode;
	int isTheFirstSession = 0;
	struct stat finfo;
	FILE *fd1;
	char *cp;
	char bufTmp[BUF_TMP_SIZE];
	char buffer[BUF_TMP_SIZE];

	// Get informations on the autolock pid file
	rCode = lstat(PID_AUTOLOCK_FILE, &finfo);

	if ((rCode == -1) && (errno == ENOENT))
	{
		// The autolock pid file doesn't exist, it's the first session
		isTheFirstSession = 1;
	}

	else if ((rCode == -1) && (errno != ENOENT))
	{
		pam_log(AUTOLOCK_IDENT, LOG_ERR,
			"Can't stat the autolock pid file: %s", strerror(errno));
		exit(EXIT_FAILURE);
	}

	else if (finfo.st_size == 0)
	{
		// It's the first session
		isTheFirstSession = 1;
	}

	else
	{
		// Is the pid write in this autolock pid file correspond to
		// a running autolock daemon?
		///////////////////////////////////////////////////////////

		// Retrieve the autolock daemon pid stored in the autolock pid file
		rCode = getAutolockPid(&pid);

		if (rCode == EXIT_FAILURE)
		{
			pam_log(AUTOLOCK_IDENT, LOG_ERR,
				"UNABLE to get the autolock daemon pid stored in the autolock pid file");
			exit(EXIT_FAILURE);
		}

		// If the process having this pid exists, its name is stored in
		// /proc/pid/cmdline
		///////////////////////////////////////////////////////////////

		sprintf(bufTmp, "/proc/%d/cmdline", pid);

		// Open the /proc/ file
		fd1 = fopen(bufTmp, "r");

		if (fd1 != NULL)
		{
			// A process having this pid exists.
			// We have to parse the cmdline to get its name.
			////////////////////////////////////////////////

			// Store the cmdline in a buffer
			fgets(buffer, MAX_LINE_LENGTH, fd1);

			if (buffer == NULL)
			{
				pam_log(AUTOLOCK_IDENT, LOG_ERR,
					"UNABLE to read in %s file: %s", bufTmp, strerror(errno));
				exit(EXIT_FAILURE);
			}

			fclose(fd1);

			// Search the last / of the command line
			cp = strrchr(buffer, '/');

			// Now, we are positionned on the name.

			if (cp == NULL)
			{
				if (strcmp(buffer, AUTOLOCK) != 0)
					isTheFirstSession = 1;
				// This pid dont correspond to an autolock process
			}
		}
		else
		{
			// This pid doesn't correspond to an alive process
			isTheFirstSession = 1;
		}
	}

	if (isTheFirstSession)
	{
		// Create the autolock pid file which contain 
		// the pid of autolock and the pid of the login process
		// opening session and initialize the opened session pid list
		/////////////////////////////////////////////////////////////

		rCode = listInit(userName);

		if (rCode == EXIT_FAILURE)
		{
			pam_log(AUTOLOCK_IDENT, LOG_ERR,
				"UNABLE to initialize the autolock pid file");
			exit(EXIT_FAILURE);
		}
	}

	if (rCode == EXIT_FAILURE)
	{
		pam_log(AUTOLOCK_IDENT, LOG_ERR,
			"UNABLE to add this session in the autolock pid file");
		exit(EXIT_FAILURE);
	}

	if (isTheFirstSession)
	{
		// We can finally call the autolock loop
		/////////////////////////////////////////

		pam_log(AUTOLOCK_IDENT, LOG_INFO, "Autolock daemon is running now");

		autolock_loop(userName, csnRequiredTxt, sessionPid);
	}
	else
	{
		// Add the pid of this new session to the autolock pid file
		rCode = putPidInList(sessionPid);
	}

	exit(EXIT_SUCCESS);
}

void remove_session(int sessionPid)
{
	// Removes a registred session from the list and stops the
	// daemon if it's the last one
	///////////////////////////////////////////////////////////

	int rCode;
	int result;

	// First, the pid of the sessionprocess must be 
	// registred in the pid list.
	///////////////////////////////////////////////

	rCode = pidInList(sessionPid, &result);

	if (rCode == EXIT_FAILURE)
	{
		pam_log(AUTOLOCK_IDENT, LOG_ERR,
			"UNABLE to search this session pid in the autolock pid file");
		exit(EXIT_FAILURE);
	}

	if (!result)
	{
		// If can be possible with xdm... The process calling
		// the open session part (which pid is registred on the list
		// is not the same of the process calling the close session part.

		pam_log(AUTOLOCK_IDENT, LOG_ERR,
			"The session %d wasn't registred in the autolock pid file",
			sessionPid);

		rCode = cleanupList();

		if (rCode == EXIT_FAILURE)
		{
			pam_log(AUTOLOCK_IDENT, LOG_ERR,
				"UNABLE to clean the autolock pid file");
			exit(EXIT_FAILURE);
		}
	}
	else
	{
		// Delete the session in the autolock pid file
		if (deletePidInList(sessionPid) == EXIT_FAILURE)
		{
			pam_log(AUTOLOCK_IDENT, LOG_ERR,
				"UNABLE to delete a session pid in the autolock pid file");
			exit(EXIT_FAILURE);
		}
	}

	// Is the session pid list empty?
	rCode = listEmpty(&result);

	if (rCode == EXIT_FAILURE)
		exit(EXIT_FAILURE);

	if (result)
	{
		// It's the last session, we can stop the autolock daemon
		rCode = stopAutolock();

		if (rCode == EXIT_FAILURE)
		{
			pam_log(AUTOLOCK_IDENT, LOG_ALERT,
				"UNABLE to stop the autolock daemon");
			exit(EXIT_FAILURE);
		}
	}

	exit(EXIT_SUCCESS);
}

void autolock_loop(char *userName, char *csnRequiredTxt, int sessionPid)
{
	// This is the main loop which tests the presence of the
	// authenticated card in the reader and locks the
	// virtuals console otherwise
	////////////////////////////////////////////////////////

	int rCode, pid, result;
#if 0
	struct passwd *structPasswd;
#endif

	// We have to fork
	pid = fork();

	if (pid == -1)
	{
		pam_log(AUTOLOCK_IDENT, LOG_ERR, "Unable to fork");
		exit(EXIT_FAILURE);
	}
	else if (pid != 0)
	{
		// We are into the father
		// just return to the pam module
		exit(EXIT_SUCCESS);
	}

	// We are on the child
	//////////////////////

	// Detach the process from the terminal;
	setsid();

	// Change the uid to the user uid in order to
	// The daemon runs under the user name
	/////////////////////////////////////////////

	// Write the daemon pid in the autolock pid file
	setAutolockPid(getpid());

	// Add the pid of the first session to 
	// the autolock pid file
	rCode = putPidInList(sessionPid);

	// change #ifdef for login(1) or gdm(1)
	// It works as root for kdm
#if 0
	// Get the user uid in the passwd file
	structPasswd = getpwnam(userName);

	if (structPasswd == NULL)
	{
		pam_log(AUTOLOCK_IDENT, LOG_ERR,
			"UNABLE to find the user in the passwd file: %s",
			strerror(errno));
		exit(EXIT_FAILURE);
	}

	// Change the saved uid
	rCode = setuid(structPasswd->pw_uid);

	if (rCode == -1)
	{
		pam_log(AUTOLOCK_IDENT, LOG_ERR,
			"UNABLE to set the autolockd daemon uid to user id: %s",
			strerror(errno));
		exit(EXIT_FAILURE);
	}
#endif

	// Install the handler of the signals
	/////////////////////////////////////

	signal(SIGUSR1, sig_usr1_handler);
	signal(SIGUSR2, sig_usr2_handler);
	signal(SIGQUIT, sig_quit_handler);

	for (;;)
	{
		// The main loop, which test permanently if the authenticated card
		// is in the reader.
		//////////////////////////////////////////////////////////////////

		if (!Suspended)
		{
			if (is_good_card_in_reader(csnRequiredTxt) == EXIT_FAILURE)
			{
				// No card in the reader or not the required card
				/////////////////////////////////////////////////

				// Here we check the autolock pid file in order to verify
				// that all the registred sessions are still alive.
				// It's possible that a session has been destroyed hardly
				// and hasn't could remove his pid from the autolock pid file
				/////////////////////////////////////////////////////////////
				rCode = cleanupList();

				if (rCode == EXIT_FAILURE)
				{
					pam_log(AUTOLOCK_IDENT, LOG_ERR,
						"UNABLE to clean the autolock pid file");
					exit(EXIT_FAILURE);
				}

				// Now, is the session pid list empty?
				rCode = listEmpty(&result);

				if (rCode == EXIT_FAILURE)
					exit(EXIT_FAILURE);

				if (result)
				{
					// It's the last session, we can stop the autolock daemon
					rCode = stopAutolock();

					if (rCode == EXIT_FAILURE)
					{
						pam_log(AUTOLOCK_IDENT, LOG_ALERT,
							"UNABLE to properly stop the autolock daemon");
						exit(EXIT_FAILURE);
					}
				}

				if ((pid = fork()) == 0)
				{
					// Execute the console locker (smartlocker).
					// We have to exec it on a new console in order to get 
					// the terminal. 
					//////////////////////////////////////////////////////

					rCode = execlp("openvt", "openvt", "-s", "-w", "--",
						"su", "-c", SMARTLOCKER_LOCATION, userName, NULL);

					if (rCode == -1)
					{
						pam_log(AUTOLOCK_IDENT, LOG_ALERT,
							"UNABLE to run smartlocker: %s", strerror(errno));
						exit(EXIT_FAILURE);
					}
				}
				else
				{
					// Wait for the end of smartlocker
					waitpid(pid, NULL, 0);
				}
			}
		}
		// Wait a little before the next test
		sleep(AUTOLOCK_LOOP_DELAY);
	}
	// We should really return something...
	exit(EXIT_SUCCESS);
}


int stopAutolock(void)
{
	// This function stop autolock and wait until 
	// the user pick up his card
	////////////////////////////////////////////

	int rCode;
	int pid;

	// Retrieve the autolock daemon pid
	rCode = getAutolockPid(&pid);

	if (rCode == EXIT_FAILURE)
	{
		pam_log(AUTOLOCK_IDENT, LOG_ERR,
			"UNABLE to get the autolock daemon pid");
		return (EXIT_FAILURE);
	}

	pam_log(AUTOLOCK_IDENT, LOG_INFO, "Killing the autolock daemon: %d", pid);

	// Kill the autolock daemon by sending the SIGQUIT signal
	rCode = kill(pid, SIGQUIT);

	if (rCode == -1)
	{
		pam_log(AUTOLOCK_IDENT, LOG_ERR,
			"UNABLE to send the QUIT signal to the autolock daemon: %s",
			strerror(errno));
		return (EXIT_FAILURE);
	}

	pam_log(AUTOLOCK_IDENT, LOG_INFO, "signal SIGQUIT sent to %d", pid);

	// we are waiting for ourself
	if (pid == getpid())
		return EXIT_SUCCESS;

	// wait until the "master" autolock exists
	while (rCode == 0)
	{
		rCode = kill(pid, 0);
		usleep(TIME_BETWEEN_BELL*1000);
	}

	return (EXIT_SUCCESS);
}

