/*
 *
 * This program is written by
 * Vincent Guerin
 * and Ludovic Rousseau <ludovic.rousseau@gemplus.com>
 *
 * $Id: smartlocker.c,v 1.5 2002/03/07 10:33:05 rousseau Exp $
 *
 */

/*
 * Copyright  2000 Gemplus
 * Av. du Pic de Bertagne - Parc d'Activits de Gmenos
 * BP 100 - 13881 Gmenos CEDEX
 * 
 * Everyone is allowed to redistribute and use this source  (source
 * code)  and binary (object code),  with or  without modification,
 * under some conditions:
 * 
 *  - Everyone  must  retain  and/or  reproduce the above copyright
 *    notice,  and the below  disclaimer of warranty and limitation
 *    of liability  for redistribution and use of these source code
 *    and object code.
 * 
 *  - Everyone  must  ask a  specific prior written permission from
 *    Gemplus to use the name of Gemplus.
 * 
 * DISCLAIMER OF WARRANTY
 * 
 * THIS CODE IS PROVIDED "AS IS",  WITHOUT ANY WARRANTY OF ANY KIND
 * (INCLUDING,  BUT  NOT  LIMITED  TO,  THE IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY  AND FITNESS FOR  A  PARTICULAR PURPOSE)  EITHER
 * EXPRESS OR IMPLIED.  GEMPLUS DOES NOT WARRANT THAT THE FUNCTIONS
 * CONTAINED  IN THIS SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR
 * THAT THE OPERATION OF IT WILL BE UNINTERRUPTED OR ERROR-FREE. NO
 * USE  OF  ANY  CODE  IS  AUTHORIZED  HEREUNDER EXCEPT UNDER  THIS
 * DISCLAIMER.
 * 
 * LIMITATION OF LIABILITY
 * 
 * GEMPLUS SHALL NOT BE LIABLE FOR INFRINGEMENTS OF  THIRD  PARTIES
 * RIGHTS. IN NO EVENTS, UNLESS REQUIRED BY APPLICABLE  LAW,  SHALL
 * GEMPLUS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES OF ANY CHARACTER  INCLUDING,
 * WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK STOPPAGE,
 * COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER DAMAGES OR
 * LOSSES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. ALSO,
 * GEMPLUS IS  UNDER NO  OBLIGATION TO MAINTAIN,  CORRECT,  UPDATE,
 * CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS SOFTWARE.
 */
/*
 * This program is copyright (C) 1994 Michael K. Johnson, and is free
 * software which is freely distributable under the terms of the
 * GNU public license, included as the file COPYING in this
 * distribution.  It is NOT public domain software, and any
 * redistribution not permitted by the GNU General Public License is
 * expressly forbidden without prior written permission from
 * the author.
 *
 */

/*
 * This program is based on the lockvc program:
 * 
 * This program is copyright (C) 1994 Michael K. Johnson, and is free
 * software which is freely distributable under the terms of the
 * GNU public license, included as the file COPYING in this
 * distribution.  It is NOT public domain software, and any
 * redistribution not permitted by the GNU General Public License is
 * expressly forbidden without prior written permission from
 * the author.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <getopt.h>
#include <termios.h>
#include <signal.h>
#include <time.h>
#include <pwd.h>
#include <sys/vt.h>
#include <sys/kd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <security/pam_appl.h>
#include <security/pam_misc.h>

static struct pam_conv PAM_conversation = {
	&misc_conv,
	NULL
};

void release_vt(int signo);
void acquire_vt(int signo);
void set_signal_mask(int save);
void signal_ignorer(int signo);
void get_password(char *username);
int correct_password(char *username);

/* for both mask_signals() and restore_signals() */
static sigset_t osig;

/* Copy of the VT mode when the program was started */
struct vt_mode ovtm;
int vfd;
struct termios oterm;
unsigned char lines, columns;
void *screen_buf = NULL;
int vcs;

int main(int argc, char **argv)
{
	struct vt_mode vtm;
	struct passwd *pw;
	pam_handle_t *pamh;
	int pam_error;
	char username[40];			/* current user's name */

	/* Execute the close session PAM modules */

	pam_error = pam_start("smartlocker", username, &PAM_conversation, &pamh);

	if (pam_error == PAM_SUCCESS)
	{
		pam_error = pam_close_session(pamh, 0);

		if (pam_error != PAM_SUCCESS)
		{
			fprintf(stderr,
				"Unable to execute the PAM close session modules\n");
			return (1);
		}

		pam_end(pamh, PAM_SUCCESS);
	}
	else
	{
		fprintf(stderr, "Unable to start PAM\n");
		return (1);
	}

	/* Get the user's passwords.  */
	pw = getpwuid(getuid());
	strncpy(username, pw->pw_name, sizeof(username) - 1);
	username[sizeof(username) - 1] = '\0';

	if ((vfd = open("/dev/tty", O_RDWR)) < 0)
	{
		perror("smartlocker: could not open /dev/tty");
		exit(1);
	}

	/* Now set the signals so we can't be summarily executed or stopped, */
	/* and handle SIGUSR{1,2} and SIGCHLD */

	set_signal_mask(1);

	ovtm = vtm;					/* Keep a copy around to restore at appropriate times */
	vtm.mode = VT_PROCESS;
	vtm.relsig = SIGUSR1;		/* handled by release_vt() */
	vtm.acqsig = SIGUSR2;		/* handled by acquire_vt() */
	ioctl(vfd, VT_SETMODE, &vtm);

	/* get_password() sets the terminal characteristics and does not */
	/* return until the correct password has been read.              */
	get_password(username);

	/* vt status was restored in restore_terminal() already... */
	if (screen_buf)
	{
		lseek(vcs, 0, SEEK_SET);
		write(vcs, &columns, 1);
		write(vcs, &lines, 1);
		write(vcs, screen_buf, 2 * lines * columns + 2);
		close(vcs);
	}

	/* we should really return something... */
	return (0);
}

int correct_password(char *username)
{
	int userAuthenticated = 0;
	pam_handle_t *pamh;
	int pam_error;

	/* Now use PAM to do authentication. */
	/* and to execute the open session modules */

	pam_error = pam_start("smartlocker", username, &PAM_conversation, &pamh);

	if (pam_error != PAM_SUCCESS)
	{
		/* fix signals that may have been disordered by pam */
		set_signal_mask(0);
		fprintf(stderr, "Unable to start PAM\n");
		return 0;
	}

	pam_error = pam_authenticate(pamh, 0);

	if (pam_error == PAM_SUCCESS)
	{
		userAuthenticated = 1;

		/* Execute the open session PAM modules */
		pam_error = pam_open_session(pamh, 0);

		if (pam_error != PAM_SUCCESS)
		{
			fprintf(stderr,
				"Unable to execute the PAM open session modules\n");
			return (1);
		}

		/* fix signals that may have been disordered by pam */
		set_signal_mask(0);
		pam_end(pamh, PAM_SUCCESS);
	}

	return (userAuthenticated);
}

void get_password(char *username)
{
	static struct termios term;
	static int already = 0;
	time_t theTime;

	/* To do: allow logging the user out safely without shutting down
	   the whole machine...  */

	write(1, "\33[H\33[J", 6);	/* There's no need to use ncurses */

	time(&theTime);

	printf("The entire console display is now completely locked.\n");
	printf("You will not be able to switch to another virtual console.\n");
	printf("\nLocked by %s on %s\n", username, ctime(&theTime));

	if (!already)
	{
		tcgetattr(STDIN_FILENO, &oterm);
		term = oterm;
		term.c_iflag &= ~BRKINT;
		term.c_iflag |= IGNBRK;
		term.c_lflag &= ~ISIG;
		already = 1;
	}

	term.c_lflag &= ~(ECHO | ECHOCTL);
	tcsetattr(STDIN_FILENO, TCSANOW, &term);

	do
	{
		/* correct_password() sets the terminal status as necessary */

		if (correct_password(username))
		{
			printf("Authentication OK\n");

			sigprocmask(SIG_SETMASK, &osig, NULL);
			ioctl(vfd, VT_SETMODE, &ovtm);
			tcsetattr(STDIN_FILENO, TCSANOW, &oterm);
			return;
		}

		/* correct_password() may have failed to return success because the   */
		/* terminal is closed.  In this case, it would not be appropriate to  */
		/* try again...                                                       */
		if (isatty(STDIN_FILENO) == 0)
		{
			perror("isatty");
			ioctl(vfd, VT_SETMODE, &ovtm);
			tcsetattr(STDIN_FILENO, TCSANOW, &oterm);
			exit(1);
		}

		/* Need to slow down people who are trying to break in by brute force */
		/* Note that it is technically possible to break this, but I can't    */
		/* make it happen, even knowing the code and knowing how to try.      */
		/* If you manage to kill vlock from the terminal while in this code,  */
		/* please tell me how you did it.                                     */

		printf(" *** \nAuthentication failure; please try again. *** \n");
		printf("\n");
	}
	while (1);
}

/* In release_vt() and acquire_vt(), anything which is done in
 * release_vt() must be undone in acquire_vt().  Right now, that's
 * not much...
 */

/* This is called by a signal whenever a user tries to change the VC
   with a ALT-Fn key */
void release_vt(int signo)
{
	ioctl(vfd, VT_RELDISP, 0);	/* kernel is not allowed to switch */
}

/* This is called whenever a user switches to that VC */
void acquire_vt(int signo)
{
	/* This call is not currently required under Linux, but it won't hurt,
	   either... */
	ioctl(vfd, VT_RELDISP, VT_ACKACQ);
}

void signal_ignorer(int signo)
{
	return;
}

void set_signal_mask(int save)
{
	static sigset_t sig;
	static struct sigaction sa;

	/* We don't want to get any signals we don't have to, but there are   */
	/* several we must get.  I don't know whether to take the current     */
	/* signal mask and change it or to do a sigfillset and go from there. */
	/* The code should handle either, I think. */

	sigprocmask(SIG_SETMASK, NULL, &sig);
	sigdelset(&sig, SIGUSR1);
	sigdelset(&sig, SIGUSR2);
	sigaddset(&sig, SIGTSTP);
	sigaddset(&sig, SIGTTIN);
	sigaddset(&sig, SIGTTOU);
	sigaddset(&sig, SIGHUP);
	sigaddset(&sig, SIGCHLD);
	sigaddset(&sig, SIGQUIT);
	sigaddset(&sig, SIGINT);

	if (save)
		sigprocmask(SIG_SETMASK, &sig, &osig);
	else
		sigprocmask(SIG_SETMASK, &sig, NULL);

	/* we set SIGUSR{1,2} to point to *_vt() above */
	sigemptyset(&(sa.sa_mask));
	sa.sa_flags = SA_RESTART;
	sa.sa_handler = release_vt;
	sigaction(SIGUSR1, &sa, NULL);
	sa.sa_handler = acquire_vt;
	sigaction(SIGUSR2, &sa, NULL);

	/* Need to handle some signals so that we don't get killed by them */
	sa.sa_handler = signal_ignorer;
	sigaction(SIGHUP, &sa, NULL);
	sigaction(SIGQUIT, &sa, NULL);
	sigaction(SIGINT, &sa, NULL);
	sigaction(SIGTSTP, &sa, NULL);
}
