/*
 *
 * This program is written by
 * Vincent Guerin
 * and Ludovic Rousseau <ludovic.rousseau@gemplus.com>
 *
 * $Id: pam_cfs_SC.c,v 1.2 2001/01/03 09:57:23 rousseau Exp $
 *
 */

/*
 * Copyright  2000 Gemplus
 * Av. du Pic de Bertagne - Parc d'Activits de Gmenos
 * BP 100 - 13881 Gmenos CEDEX
 * 
 * Everyone is allowed to redistribute and use this source  (source
 * code)  and binary (object code),  with or  without modification,
 * under some conditions:
 * 
 *  - Everyone  must  retain  and/or  reproduce the above copyright
 *    notice,  and the below  disclaimer of warranty and limitation
 *    of liability  for redistribution and use of these source code
 *    and object code.
 * 
 *  - Everyone  must  ask a  specific prior written permission from
 *    Gemplus to use the name of Gemplus.
 * 
 * DISCLAIMER OF WARRANTY
 * 
 * THIS CODE IS PROVIDED "AS IS",  WITHOUT ANY WARRANTY OF ANY KIND
 * (INCLUDING,  BUT  NOT  LIMITED  TO,  THE IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY  AND FITNESS FOR  A  PARTICULAR PURPOSE)  EITHER
 * EXPRESS OR IMPLIED.  GEMPLUS DOES NOT WARRANT THAT THE FUNCTIONS
 * CONTAINED  IN THIS SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR
 * THAT THE OPERATION OF IT WILL BE UNINTERRUPTED OR ERROR-FREE. NO
 * USE  OF  ANY  CODE  IS  AUTHORIZED  HEREUNDER EXCEPT UNDER  THIS
 * DISCLAIMER.
 * 
 * LIMITATION OF LIABILITY
 * 
 * GEMPLUS SHALL NOT BE LIABLE FOR INFRINGEMENTS OF  THIRD  PARTIES
 * RIGHTS. IN NO EVENTS, UNLESS REQUIRED BY APPLICABLE  LAW,  SHALL
 * GEMPLUS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES OF ANY CHARACTER  INCLUDING,
 * WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK STOPPAGE,
 * COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER DAMAGES OR
 * LOSSES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. ALSO,
 * GEMPLUS IS  UNDER NO  OBLIGATION TO MAINTAIN,  CORRECT,  UPDATE,
 * CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS SOFTWARE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <security/pam_modules.h>

#include "config.h"
#include "global.h"
#include "pam_smartcard_utils.h"

// PAM services offered
#define PAM_SM_SESSION

// This module provides only the session part. 
// 
// The open session function calls the cattach_SC binary with the -a
// option in order to attach the ciphered directries of the user cfs
// fstab whose auto attach flag is set to auto.
// This function try to get the user PIN code from the pam handle
// structure; if it's not present, it uses the pam conversate machanism
// to tell it to the user.
//
// The close session function calls the cdetach_SC binary with the -a
// option in order to detach all the currently attached directories
// (There are registred in the user cfs mtbab file).
///////////////////////////////////////////////////////////////////////

// Internal cleaning up function needed by the data mechanism
void cleanup_func(pam_handle_t * pamh, void *data, int error_status)
{
	// This data is the usr PIN code
	// To obtain maximum of security we overwrite the memory with 0
	memset(data, USR_PIN_SIZE, '\0');
}

PAM_EXTERN int pam_sm_open_session(pam_handle_t * pamh, int flags,
	int argc, const char **argv)
{
	int rCode;
	char *userName;
	char *pin = NULL;
	char bufTmp[BUF_TMP_SIZE];

	// First get the user name
	rCode = pam_get_user(pamh, (const char **) &userName, NULL);

	if ((rCode != PAM_SUCCESS) || (!userName))
	{
		pam_log(PAM_CFS_SC, LOG_ERR, "Unable to get the user name");
		return (PAM_SESSION_ERR);
	}

	// Perhaps the pin is already present in the pam handle sructure
	// This module will be generally used in concordance with the 
	// pam_smartcard authentication module, which is designed to
	// store the PIN in the PAM structure.
	////////////////////////////////////////////////////////////////

	rCode = pam_get_data(pamh, PAM_PIN_DATA_NAME, (void *) &pin);

	if (rCode != PAM_SUCCESS)
	{
		// Doesn't matter, tell to the user that he have to insert his card
		// Wait for it and... and for his PIN code...

		pin = calloc(USR_PIN_SIZE + 1, sizeof(char));

		// Remark that we use the same function as for the pam smartcard module
		rCode = pam_conversate(pamh, pin);

		if (rCode != PAM_SUCCESS)
		{
			pam_log(PAM_CFS_SC, LOG_CRIT, "Conversate failed");
			return (PAM_SESSION_ERR);
		}

		// Put the PIN code in the pamh handle structure
		// to provide it to hypothetical other modules
		// Actually, I don't think that such modules exist
		// but furthermore I hope so.
		//////////////////////////////////////////////////

		pam_set_data(pamh, PAM_PIN_DATA_NAME, pin, cleanup_func);
	}

	// Now, we can call the cattach_SC function with these options:
	// -p pin: to directly give the user PIN to the binary;
	// -a: indicates that all the directories whose the auto_attach 
	// flag is set to auto must be attached;
	// -q: indicates that the module must be quiet and so doesn't display
	// any messages.
	////////////////////////////////////////////////////////////////////

	if (getuid() == 0)
	{
		// If caller is root, we must use "su" to execute "cattach_SC" 
		// under the user name
		//////////////////////////////////////////////////////////////

		sprintf(bufTmp, "/bin/su %s -c \"%s -p %s -a -q\"", userName,
			CFS_SC_BIN_DIR CATTACH_SC, pin);
	}
	else
		sprintf(bufTmp, "%s -p %s -a -q", CFS_SC_BIN_DIR CATTACH_SC, pin);

	rCode = system(bufTmp);

	if (rCode == 127)
	{
		pam_log(PAM_CFS_SC, LOG_ERR,
			"Unable to execute the binary %s: %s", CATTACH_SC,
			strerror(errno));
		return (PAM_SESSION_ERR);
	}

	return (PAM_SUCCESS);
}

PAM_EXTERN int pam_sm_close_session(pam_handle_t * pamh, int flags,
	int argc, const char **argv)
{
	int rCode;
	char *userName;
	char bufTmp[BUF_TMP_SIZE];

	// First get the user name
	rCode = pam_get_user(pamh, (const char **) &userName, NULL);

	if ((rCode != PAM_SUCCESS) || (!userName))
	{
		pam_log(PAM_CFS_SC, LOG_ERR, "Unable to get the user name");
		return (PAM_SESSION_ERR);
	}

	// Just call the cdetach_SC function with these options:
	// -a: indicates that all the directories currently attached
	// must be detached.
	// -q: indicates that the module must be quiet and so doesn't 
	// display any messages.
	/////////////////////////////////////////////////////////////

	if (getuid() == 0)
	{
		// If caller is root, we must use su to execute "cdetach_SC" 
		// under the user name
		////////////////////////////////////////////////////////////

		sprintf(bufTmp, "/bin/su %s -c \"%s -a -q\"",
			userName, CFS_SC_BIN_DIR CDETACH_SC);
	}
	else
		sprintf(bufTmp, "%s -a -q", CFS_SC_BIN_DIR CDETACH_SC);

	// Now, we can execute the command
	rCode = system(bufTmp);

	if (rCode == 127)
	{
		pam_log(PAM_CFS_SC, LOG_ERR,
			"Unable to execute the binary %s: %s", CDETACH_SC,
			strerror(errno));
		exit(EXIT_FAILURE);
	}

	return (PAM_SUCCESS);
}

#ifdef PAM_STATIC
// Static module data 
struct pam_module _pam_smartcard_modstruct = {
	"pam_smartcard",
	NULL,
	NULL,
	NULL,
	pam_sm_open_session,
	pam_sm_close_session,
	NULL
};
#endif
