/*
 *
 * This program is written by
 * Vincent Guerin
 * and Ludovic Rousseau <ludovic.rousseau@gemplus.com>
 *
 * $Id: chpin.c,v 1.2 2001/01/03 09:58:01 rousseau Exp $
 *
 */

/*
 * Copyright  2000 Gemplus
 * Av. du Pic de Bertagne - Parc d'Activits de Gmenos
 * BP 100 - 13881 Gmenos CEDEX
 * 
 * Everyone is allowed to redistribute and use this source  (source
 * code)  and binary (object code),  with or  without modification,
 * under some conditions:
 * 
 *  - Everyone  must  retain  and/or  reproduce the above copyright
 *    notice,  and the below  disclaimer of warranty and limitation
 *    of liability  for redistribution and use of these source code
 *    and object code.
 * 
 *  - Everyone  must  ask a  specific prior written permission from
 *    Gemplus to use the name of Gemplus.
 * 
 * DISCLAIMER OF WARRANTY
 * 
 * THIS CODE IS PROVIDED "AS IS",  WITHOUT ANY WARRANTY OF ANY KIND
 * (INCLUDING,  BUT  NOT  LIMITED  TO,  THE IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY  AND FITNESS FOR  A  PARTICULAR PURPOSE)  EITHER
 * EXPRESS OR IMPLIED.  GEMPLUS DOES NOT WARRANT THAT THE FUNCTIONS
 * CONTAINED  IN THIS SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR
 * THAT THE OPERATION OF IT WILL BE UNINTERRUPTED OR ERROR-FREE. NO
 * USE  OF  ANY  CODE  IS  AUTHORIZED  HEREUNDER EXCEPT UNDER  THIS
 * DISCLAIMER.
 * 
 * LIMITATION OF LIABILITY
 * 
 * GEMPLUS SHALL NOT BE LIABLE FOR INFRINGEMENTS OF  THIRD  PARTIES
 * RIGHTS. IN NO EVENTS, UNLESS REQUIRED BY APPLICABLE  LAW,  SHALL
 * GEMPLUS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES OF ANY CHARACTER  INCLUDING,
 * WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK STOPPAGE,
 * COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER DAMAGES OR
 * LOSSES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. ALSO,
 * GEMPLUS IS  UNDER NO  OBLIGATION TO MAINTAIN,  CORRECT,  UPDATE,
 * CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS SOFTWARE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "config.h"
#include "global.h"
#include "smartcard_functions.h"

// Just an interface for the change pin function of the 
// smart_card_functions library
//
// It allows to the user the change his PIN code hmself.

// User must give the current PIN code (which will be 
// verify) and the new PIN code (twice to be sure of it)
////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
	int rCode;
	char *pinReaded;
	char newPin1[USR_PIN_SIZE + 2];
	char newPin2[USR_PIN_SIZE + 2];
	char oldPin[USR_PIN_SIZE + 2];

	if (argc != 1)
	{
		// No arguments required
		printf("%s : change the user card pin code\n", argv[0]);
		printf("<vincent.guerin@gemplus.com>, 27/10/2000\n");
		exit(EXIT_FAILURE);
	}

	// The PIN codes are get using the getpass function
	// in order to switch off the echo, increasing the
	// security.
	// We store under USR_PIN_SIZE + 1 character in order 
	// to detect an error when the given PIN is too long.
	/////////////////////////////////////////////////////

	// Get the current PIN
	pinReaded = getpass("Enter the current pin: ");
	strncpy(oldPin, pinReaded, USR_PIN_SIZE + 1);
	oldPin[USR_PIN_SIZE + 1] = '\0';

	// Get the new PIN
	pinReaded = getpass("Enter the new pin: ");
	strncpy(newPin1, pinReaded, USR_PIN_SIZE + 1);
	newPin1[USR_PIN_SIZE + 1] = '\0';

	// Get the new PIN twice
	pinReaded = getpass("Re-Enter the new pin: ");
	strncpy(newPin2, pinReaded, USR_PIN_SIZE + 1);
	newPin2[USR_PIN_SIZE + 1] = '\0';

	// The two new PINs must be the same
	if (strcmp(newPin1, newPin2) != 0)
	{
		fprintf(stderr, "ERROR: PINs MISMATCH, try again!!!\n");
		exit(EXIT_FAILURE);
	}

	// Test the size of the new PIN
	if (strlen(newPin1) != USR_PIN_SIZE)
	{
		fprintf(stderr, "ERROR: incorect PIN size, %d characters required!\n",
			USR_PIN_SIZE);
		exit(EXIT_FAILURE);
	}

	// Call the change pin function of the smartcard_functions library
	rCode = change_pin(oldPin, newPin1);

	if (rCode == ERROR_CARD_NOT_FORMATED)
	{
		// The card is probably not formatted for the smartsession applications
		fprintf(stderr, "ERROR: this card seems to be NOT FORMATTED!\n");
		exit(EXIT_FAILURE);
	}

	if (rCode == ERROR_VERIFY_PIN)
	{
		// The current PIN given by the user is wrong
		fprintf(stderr, "ERROR: current PIN code is WRONG!\n");
		exit(EXIT_FAILURE);
	}

	exit(EXIT_SUCCESS);
}
